import copy

import pytest

import salt.utils.json
import salt.utils.schema as schema
import salt.utils.stringutils
import salt.utils.yaml
from salt.utils.versions import Version
from tests.support.unit import TestCase

try:
    import jsonschema
    import jsonschema.exceptions

    HAS_JSONSCHEMA = True
    JSONSCHEMA_VERSION = Version(jsonschema.__version__)
except ImportError:
    HAS_JSONSCHEMA = False
    JSONSCHEMA_VERSION = Version("0")


# pylint: disable=unused-import,function-redefined
try:
    import rfc3987

    HAS_RFC3987 = True
except ImportError:
    HAS_RFC3987 = False

try:
    import strict_rfc3339

    HAS_STRICT_RFC3339 = True
except ImportError:
    HAS_STRICT_RFC3339 = False
# pylint: enable=unused-import


class ConfigTestCase(TestCase):
    """
    TestCase for salt.utils.config module
    """

    def test_configuration_subclass_inherits_items(self):
        class BaseConfig(schema.Schema):
            base = schema.BooleanItem(default=True, required=True)

        class SubClassedConfig(BaseConfig):
            hungry = schema.BooleanItem(
                title="Hungry", description="Are you hungry?", required=True
            )

        self.assertDictEqual(
            SubClassedConfig.serialize(),
            {
                "$schema": "http://json-schema.org/draft-04/schema#",
                "type": "object",
                "properties": {
                    "base": {"default": True, "type": "boolean", "title": "base"},
                    "hungry": {
                        "type": "boolean",
                        "description": "Are you hungry?",
                        "title": "Hungry",
                    },
                },
                "required": ["base", "hungry"],
                "x-ordering": ["base", "hungry"],
                "additionalProperties": False,
            },
        )

        class MergedConfigClass(schema.Schema):
            thirsty = schema.BooleanItem(
                title="Thirsty", description="Are you thirsty?", required=True
            )
            merge_subclassed = SubClassedConfig(flatten=True)

        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "properties": {
                "thirsty": {
                    "type": "boolean",
                    "description": "Are you thirsty?",
                    "title": "Thirsty",
                },
                "base": {"default": True, "type": "boolean", "title": "base"},
                "hungry": {
                    "type": "boolean",
                    "description": "Are you hungry?",
                    "title": "Hungry",
                },
            },
            "required": ["thirsty", "base", "hungry"],
            "x-ordering": ["thirsty", "base", "hungry"],
            "additionalProperties": False,
        }
        self.assertDictContainsSubset(
            MergedConfigClass.serialize()["properties"], expected["properties"]
        )
        self.assertDictContainsSubset(expected, MergedConfigClass.serialize())

    def test_configuration_items_order(self):
        class One(schema.Schema):
            one = schema.BooleanItem()

        class Three(schema.Schema):
            three = schema.BooleanItem()

        class Final(One):
            two = schema.BooleanItem()
            three = Three(flatten=True)

        self.assertEqual(Final.serialize()["x-ordering"], ["one", "two", "three"])

    def test_optional_requirements_config(self):
        class BaseRequirements(schema.Schema):
            driver = schema.StringItem(default="digitalocean", format="hidden")

        class SSHKeyFileSchema(schema.Schema):
            ssh_key_file = schema.StringItem(
                title="SSH Private Key",
                description=(
                    "The path to an SSH private key which will be used "
                    "to authenticate on the deployed VMs"
                ),
            )

        class SSHKeyNamesSchema(schema.Schema):
            ssh_key_names = schema.StringItem(
                title="SSH Key Names",
                description=(
                    "The names of an SSH key being managed on "
                    "DigitalOcean account which will be used to "
                    "authenticate on the deployed VMs"
                ),
            )

        class Requirements(BaseRequirements):
            title = "DigitalOcean"
            description = "DigitalOcean Cloud VM configuration requirements."

            personal_access_token = schema.StringItem(
                title="Personal Access Token",
                description=(
                    "This is the API access token which can be generated "
                    "under the API/Application on your account"
                ),
                required=True,
            )

            requirements_definition = schema.AnyOfItem(
                items=(
                    SSHKeyFileSchema.as_requirements_item(),
                    SSHKeyNamesSchema.as_requirements_item(),
                ),
            )(flatten=True)
            ssh_key_file = SSHKeyFileSchema(flatten=True)
            ssh_key_names = SSHKeyNamesSchema(flatten=True)

        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "DigitalOcean",
            "description": "DigitalOcean Cloud VM configuration requirements.",
            "type": "object",
            "properties": {
                "driver": {
                    "default": "digitalocean",
                    "format": "hidden",
                    "type": "string",
                    "title": "driver",
                },
                "personal_access_token": {
                    "type": "string",
                    "description": (
                        "This is the API access token which can be "
                        "generated under the API/Application on your account"
                    ),
                    "title": "Personal Access Token",
                },
                "ssh_key_file": {
                    "type": "string",
                    "description": (
                        "The path to an SSH private key which will "
                        "be used to authenticate on the deployed VMs"
                    ),
                    "title": "SSH Private Key",
                },
                "ssh_key_names": {
                    "type": "string",
                    "description": (
                        "The names of an SSH key being managed on DigitalOcean "
                        "account which will be used to authenticate on the deployed VMs"
                    ),
                    "title": "SSH Key Names",
                },
            },
            "anyOf": [{"required": ["ssh_key_file"]}, {"required": ["ssh_key_names"]}],
            "required": ["personal_access_token"],
            "x-ordering": [
                "driver",
                "personal_access_token",
                "ssh_key_file",
                "ssh_key_names",
            ],
            "additionalProperties": False,
        }
        self.assertDictEqual(expected, Requirements.serialize())

        class Requirements2(BaseRequirements):
            title = "DigitalOcean"
            description = "DigitalOcean Cloud VM configuration requirements."

            personal_access_token = schema.StringItem(
                title="Personal Access Token",
                description=(
                    "This is the API access token which can be generated "
                    "under the API/Application on your account"
                ),
                required=True,
            )

            ssh_key_file = schema.StringItem(
                title="SSH Private Key",
                description=(
                    "The path to an SSH private key which will be used "
                    "to authenticate on the deployed VMs"
                ),
            )

            ssh_key_names = schema.StringItem(
                title="SSH Key Names",
                description=(
                    "The names of an SSH key being managed on "
                    "DigitalOcean account which will be used to "
                    "authenticate on the deployed VMs"
                ),
            )

            requirements_definition = schema.AnyOfItem(
                items=(
                    schema.RequirementsItem(requirements=["ssh_key_file"]),
                    schema.RequirementsItem(requirements=["ssh_key_names"]),
                ),
            )(flatten=True)

        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "DigitalOcean",
            "description": "DigitalOcean Cloud VM configuration requirements.",
            "type": "object",
            "properties": {
                "driver": {
                    "default": "digitalocean",
                    "format": "hidden",
                    "type": "string",
                    "title": "driver",
                },
                "personal_access_token": {
                    "type": "string",
                    "description": (
                        "This is the API access token which can be "
                        "generated under the API/Application on your account"
                    ),
                    "title": "Personal Access Token",
                },
                "ssh_key_file": {
                    "type": "string",
                    "description": (
                        "The path to an SSH private key which will "
                        "be used to authenticate on the deployed VMs"
                    ),
                    "title": "SSH Private Key",
                },
                "ssh_key_names": {
                    "type": "string",
                    "description": (
                        "The names of an SSH key being managed on DigitalOcean "
                        "account which will be used to authenticate on the deployed VMs"
                    ),
                    "title": "SSH Key Names",
                },
            },
            "anyOf": [{"required": ["ssh_key_file"]}, {"required": ["ssh_key_names"]}],
            "required": ["personal_access_token"],
            "x-ordering": [
                "driver",
                "personal_access_token",
                "ssh_key_file",
                "ssh_key_names",
            ],
            "additionalProperties": False,
        }
        self.assertDictContainsSubset(expected, Requirements2.serialize())

        class Requirements3(schema.Schema):
            title = "DigitalOcean"
            description = "DigitalOcean Cloud VM configuration requirements."

            merge_reqs = Requirements(flatten=True)

        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "DigitalOcean",
            "description": "DigitalOcean Cloud VM configuration requirements.",
            "type": "object",
            "properties": {
                "driver": {
                    "default": "digitalocean",
                    "format": "hidden",
                    "type": "string",
                    "title": "driver",
                },
                "personal_access_token": {
                    "type": "string",
                    "description": (
                        "This is the API access token which can be "
                        "generated under the API/Application on your account"
                    ),
                    "title": "Personal Access Token",
                },
                "ssh_key_file": {
                    "type": "string",
                    "description": (
                        "The path to an SSH private key which will "
                        "be used to authenticate on the deployed VMs"
                    ),
                    "title": "SSH Private Key",
                },
                "ssh_key_names": {
                    "type": "string",
                    "description": (
                        "The names of an SSH key being managed on DigitalOcean "
                        "account which will be used to authenticate on the deployed VMs"
                    ),
                    "title": "SSH Key Names",
                },
            },
            "anyOf": [{"required": ["ssh_key_file"]}, {"required": ["ssh_key_names"]}],
            "required": ["personal_access_token"],
            "x-ordering": [
                "driver",
                "personal_access_token",
                "ssh_key_file",
                "ssh_key_names",
            ],
            "additionalProperties": False,
        }
        self.assertDictContainsSubset(expected, Requirements3.serialize())

        class Requirements4(schema.Schema):
            title = "DigitalOcean"
            description = "DigitalOcean Cloud VM configuration requirements."

            merge_reqs = Requirements(flatten=True)

            ssh_key_file_2 = schema.StringItem(
                title="SSH Private Key",
                description=(
                    "The path to an SSH private key which will be used "
                    "to authenticate on the deployed VMs"
                ),
            )

            ssh_key_names_2 = schema.StringItem(
                title="SSH Key Names",
                description=(
                    "The names of an SSH key being managed on "
                    "DigitalOcean account which will be used to "
                    "authenticate on the deployed VMs"
                ),
            )

            requirements_definition_2 = schema.AnyOfItem(
                items=(
                    schema.RequirementsItem(requirements=["ssh_key_file_2"]),
                    schema.RequirementsItem(requirements=["ssh_key_names_2"]),
                ),
            )(flatten=True)

        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "DigitalOcean",
            "description": "DigitalOcean Cloud VM configuration requirements.",
            "type": "object",
            "properties": {
                "driver": {
                    "default": "digitalocean",
                    "format": "hidden",
                    "type": "string",
                    "title": "driver",
                },
                "personal_access_token": {
                    "type": "string",
                    "description": (
                        "This is the API access token which can be "
                        "generated under the API/Application on your account"
                    ),
                    "title": "Personal Access Token",
                },
                "ssh_key_file": {
                    "type": "string",
                    "description": (
                        "The path to an SSH private key which will "
                        "be used to authenticate on the deployed VMs"
                    ),
                    "title": "SSH Private Key",
                },
                "ssh_key_names": {
                    "type": "string",
                    "description": (
                        "The names of an SSH key being managed on DigitalOcean "
                        "account which will be used to authenticate on the deployed VMs"
                    ),
                    "title": "SSH Key Names",
                },
                "ssh_key_file_2": {
                    "type": "string",
                    "description": (
                        "The path to an SSH private key which will "
                        "be used to authenticate on the deployed VMs"
                    ),
                    "title": "SSH Private Key",
                },
                "ssh_key_names_2": {
                    "type": "string",
                    "description": (
                        "The names of an SSH key being managed on DigitalOcean "
                        "account which will be used to authenticate on the deployed VMs"
                    ),
                    "title": "SSH Key Names",
                },
            },
            "anyOf": [
                {"required": ["ssh_key_file"]},
                {"required": ["ssh_key_names"]},
                {"required": ["ssh_key_file_2"]},
                {"required": ["ssh_key_names_2"]},
            ],
            "required": ["personal_access_token"],
            "x-ordering": [
                "driver",
                "personal_access_token",
                "ssh_key_file",
                "ssh_key_names",
                "ssh_key_file_2",
                "ssh_key_names_2",
            ],
            "additionalProperties": False,
        }
        self.assertDictContainsSubset(expected, Requirements4.serialize())

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_optional_requirements_config_validation(self):
        class BaseRequirements(schema.Schema):
            driver = schema.StringItem(default="digitalocean", format="hidden")

        class SSHKeyFileSchema(schema.Schema):
            ssh_key_file = schema.StringItem(
                title="SSH Private Key",
                description=(
                    "The path to an SSH private key which will be used "
                    "to authenticate on the deployed VMs"
                ),
            )

        class SSHKeyNamesSchema(schema.Schema):
            ssh_key_names = schema.StringItem(
                title="SSH Key Names",
                description=(
                    "The names of an SSH key being managed on  "
                    "Digial Ocean account which will be used to "
                    "authenticate on the deployed VMs"
                ),
            )

        class Requirements(BaseRequirements):
            title = "DigitalOcean"
            description = "DigitalOcean Cloud VM configuration requirements."

            personal_access_token = schema.StringItem(
                title="Personal Access Token",
                description=(
                    "This is the API access token which can be generated "
                    "under the API/Application on your account"
                ),
                required=True,
            )

            requirements_definition = schema.AnyOfItem(
                items=(
                    SSHKeyFileSchema.as_requirements_item(),
                    SSHKeyNamesSchema.as_requirements_item(),
                ),
            )(flatten=True)
            ssh_key_file = SSHKeyFileSchema(flatten=True)
            ssh_key_names = SSHKeyNamesSchema(flatten=True)

        try:
            jsonschema.validate(
                {
                    "personal_access_token": "foo",
                    "ssh_key_names": "bar",
                    "ssh_key_file": "test",
                },
                Requirements.serialize(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        try:
            jsonschema.validate(
                {"personal_access_token": "foo", "ssh_key_names": "bar"},
                Requirements.serialize(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        try:
            jsonschema.validate(
                {"personal_access_token": "foo", "ssh_key_file": "test"},
                Requirements.serialize(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"personal_access_token": "foo"}, Requirements.serialize()
            )
        if JSONSCHEMA_VERSION >= Version("3.0.0"):
            self.assertIn(
                "'ssh_key_file' is a required property", excinfo.exception.message
            )
        else:
            self.assertIn(
                "is not valid under any of the given schemas", excinfo.exception.message
            )

    def test_boolean_config(self):
        item = schema.BooleanItem(title="Hungry", description="Are you hungry?")
        self.assertDictEqual(
            item.serialize(),
            {"type": "boolean", "title": item.title, "description": item.description},
        )

        item = schema.BooleanItem(
            title="Hungry", description="Are you hungry?", default=False
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "boolean",
                "title": item.title,
                "description": item.description,
                "default": item.default,
            },
        )

        item = schema.BooleanItem(
            title="Hungry", description="Are you hungry?", default=schema.Null
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "boolean",
                "title": item.title,
                "description": item.description,
                "default": None,
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_boolean_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.BooleanItem(title="Hungry", description="Are you hungry?")

        try:
            jsonschema.validate({"item": False}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 1}, TestConf.serialize())
        self.assertIn("is not of type", excinfo.exception.message)

    def test_string_config(self):
        item = schema.StringItem(title="Foo", description="Foo Item")
        self.assertDictEqual(
            item.serialize(),
            {"type": "string", "title": item.title, "description": item.description},
        )

        item = schema.StringItem(
            title="Foo", description="Foo Item", min_length=1, max_length=3
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "minLength": item.min_length,
                "maxLength": item.max_length,
            },
        )

        item = schema.StringItem(
            title="Foo",
            description="Foo Item",
            min_length=1,
            max_length=3,
            default="foo",
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "minLength": item.min_length,
                "maxLength": item.max_length,
                "default": "foo",
            },
        )

        item = schema.StringItem(
            title="Foo",
            description="Foo Item",
            min_length=1,
            max_length=3,
            enum=("foo", "bar"),
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "minLength": item.min_length,
                "maxLength": item.max_length,
                "enum": ["foo", "bar"],
            },
        )

        item = schema.StringItem(
            title="Foo",
            description="Foo Item",
            min_length=1,
            max_length=3,
            enum=("foo", "bar"),
            enumNames=("Foo", "Bar"),
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "minLength": item.min_length,
                "maxLength": item.max_length,
                "enum": ["foo", "bar"],
                "enumNames": ["Foo", "Bar"],
            },
        )

        item = schema.StringItem(
            title="Foo", description="Foo Item", pattern=r"^([\w_-]+)$"
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "pattern": item.pattern,
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_string_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.StringItem(title="Foo", description="Foo Item")

        try:
            jsonschema.validate({"item": "the item"}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        class TestConf(schema.Schema):
            item = schema.StringItem(
                title="Foo", description="Foo Item", min_length=1, max_length=10
            )

        try:
            jsonschema.validate({"item": "the item"}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 3}, TestConf.serialize())
        self.assertIn("is not of type", excinfo.exception.message)

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": "the item the item"}, TestConf.serialize())
        self.assertIn("is too long", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.StringItem(
                title="Foo", description="Foo Item", min_length=10, max_length=100
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": "the item"}, TestConf.serialize())
        self.assertIn("is too short", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.StringItem(
                title="Foo", description="Foo Item", enum=("foo", "bar")
            )

        try:
            jsonschema.validate({"item": "foo"}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        class TestConf(schema.Schema):
            item = schema.StringItem(
                title="Foo", description="Foo Item", enum=("foo", "bar")
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": "bin"}, TestConf.serialize())
        self.assertIn("is not one of", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.StringItem(
                title="Foo", description="Foo Item", pattern=r"^([\w_-]+)$"
            )

        try:
            jsonschema.validate(
                {"item": "the-item"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": "the item"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("does not match", excinfo.exception.message)

    def test_email_config(self):
        item = schema.EMailItem(title="Foo", description="Foo Item")
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "format": item.format,
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_email_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.EMailItem(title="Item", description="Item description")

        try:
            jsonschema.validate(
                {"item": "nobody@nowhere.com"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": "3"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is not a", excinfo.exception.message)

    def test_ipv4_config(self):
        item = schema.IPv4Item(title="Foo", description="Foo Item")
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "format": item.format,
            },
        )

    @pytest.mark.skipif(
        JSONSCHEMA_VERSION <= Version("2.5.0"),
        reason="Requires jsonschema 2.5.0 or greater",
    )
    def test_ipv4_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.IPv4Item(title="Item", description="Item description")

        try:
            jsonschema.validate(
                {"item": "127.0.0.1"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": "3"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is not a", excinfo.exception.message)

    def test_ipv6_config(self):
        item = schema.IPv6Item(title="Foo", description="Foo Item")
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "format": item.format,
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_ipv6_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.IPv6Item(title="Item", description="Item description")

        try:
            jsonschema.validate(
                {"item": salt.utils.stringutils.to_str("::1")},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": "3"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is not a", excinfo.exception.message)

    def test_hostname_config(self):
        item = schema.HostnameItem(title="Foo", description="Foo Item")
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "format": item.format,
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_hostname_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.HostnameItem(title="Item", description="Item description")

        try:
            jsonschema.validate(
                {"item": "localhost"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": "3"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is not a", excinfo.exception.message)

    def test_datetime_config(self):
        item = schema.DateTimeItem(title="Foo", description="Foo Item")
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "format": item.format,
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    @pytest.mark.skipif(
        not HAS_STRICT_RFC3339, reason="The 'strict_rfc3339' library is missing"
    )
    def test_datetime_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.DateTimeItem(title="Item", description="Item description")

        try:
            jsonschema.validate(
                {"item": "2015-07-01T18:05:27+01:00"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": "3"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is not a", excinfo.exception.message)

    def test_secret_config(self):
        item = schema.SecretItem(title="Foo", description="Foo Item")
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "format": item.format,
            },
        )

    def test_uri_config(self):
        item = schema.UriItem(title="Foo", description="Foo Item")
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "string",
                "title": item.title,
                "description": item.description,
                "format": item.format,
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    @pytest.mark.skipif(HAS_RFC3987 is False, reason="The 'rfc3987' library is missing")
    def test_uri_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.UriItem(title="Item", description="Item description")

        try:
            jsonschema.validate(
                {"item": "ssh://localhost"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": "3"},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is not a", excinfo.exception.message)

    def test_number_config(self):
        item = schema.NumberItem(title="How many dogs", description="Question")
        self.assertDictEqual(
            item.serialize(),
            {"type": "number", "title": item.title, "description": item.description},
        )

        item = schema.NumberItem(
            title="How many dogs", description="Question", minimum=0, maximum=10
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "number",
                "title": item.title,
                "description": item.description,
                "minimum": item.minimum,
                "maximum": item.maximum,
            },
        )

        item = schema.NumberItem(
            title="How many dogs", description="Question", multiple_of=2
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "number",
                "title": item.title,
                "description": item.description,
                "multipleOf": item.multiple_of,
            },
        )

        item = schema.NumberItem(
            title="How many dogs",
            description="Question",
            minimum=0,
            exclusive_minimum=True,
            maximum=10,
            exclusive_maximum=True,
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "number",
                "title": item.title,
                "description": item.description,
                "minimum": item.minimum,
                "maximum": item.maximum,
                "exclusiveMinimum": True,
                "exclusiveMaximum": True,
            },
        )

        item = schema.NumberItem(
            title="How many dogs",
            description="Question",
            minimum=0,
            maximum=10,
            default=0,
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "number",
                "title": item.title,
                "description": item.description,
                "minimum": item.minimum,
                "maximum": item.maximum,
                "default": 0,
            },
        )

        item = schema.NumberItem(
            title="How many dogs",
            description="Question",
            minimum=0,
            maximum=10,
            default=0,
            enum=(0, 2, 4, 6),
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "number",
                "title": item.title,
                "description": item.description,
                "minimum": item.minimum,
                "maximum": item.maximum,
                "default": 0,
                "enum": [0, 2, 4, 6],
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_number_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.NumberItem(title="How many dogs", description="Question")

        try:
            jsonschema.validate({"item": 2}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": "3"}, TestConf.serialize())
        self.assertIn("is not of type", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.NumberItem(
                title="How many dogs", description="Question", multiple_of=2.2
            )

        try:
            jsonschema.validate({"item": 4.4}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 4}, TestConf.serialize())
        self.assertIn("is not a multiple of", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.NumberItem(
                title="Foo", description="Foo Item", minimum=1, maximum=10
            )

        try:
            jsonschema.validate({"item": 3}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 11}, TestConf.serialize())
        self.assertIn("is greater than the maximum of", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.NumberItem(
                title="Foo", description="Foo Item", minimum=10, maximum=100
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 3}, TestConf.serialize())
        self.assertIn("is less than the minimum of", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.NumberItem(
                title="How many dogs",
                description="Question",
                minimum=0,
                exclusive_minimum=True,
                maximum=10,
                exclusive_maximum=True,
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 0}, TestConf.serialize())
        self.assertIn(
            "is less than or equal to the minimum of", excinfo.exception.message
        )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 10}, TestConf.serialize())
        self.assertIn(
            "is greater than or equal to the maximum of", excinfo.exception.message
        )

        class TestConf(schema.Schema):
            item = schema.NumberItem(
                title="Foo", description="Foo Item", enum=(0, 2, 4, 6)
            )

        try:
            jsonschema.validate({"item": 4}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        class TestConf(schema.Schema):
            item = schema.NumberItem(
                title="Foo", description="Foo Item", enum=(0, 2, 4, 6)
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 3}, TestConf.serialize())
        self.assertIn("is not one of", excinfo.exception.message)

    def test_integer_config(self):
        item = schema.IntegerItem(title="How many dogs", description="Question")
        self.assertDictEqual(
            item.serialize(),
            {"type": "integer", "title": item.title, "description": item.description},
        )

        item = schema.IntegerItem(
            title="How many dogs", description="Question", minimum=0, maximum=10
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "integer",
                "title": item.title,
                "description": item.description,
                "minimum": item.minimum,
                "maximum": item.maximum,
            },
        )

        item = schema.IntegerItem(
            title="How many dogs", description="Question", multiple_of=2
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "integer",
                "title": item.title,
                "description": item.description,
                "multipleOf": item.multiple_of,
            },
        )

        item = schema.IntegerItem(
            title="How many dogs",
            description="Question",
            minimum=0,
            exclusive_minimum=True,
            maximum=10,
            exclusive_maximum=True,
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "integer",
                "title": item.title,
                "description": item.description,
                "minimum": item.minimum,
                "maximum": item.maximum,
                "exclusiveMinimum": True,
                "exclusiveMaximum": True,
            },
        )

        item = schema.IntegerItem(
            title="How many dogs",
            description="Question",
            minimum=0,
            maximum=10,
            default=0,
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "integer",
                "title": item.title,
                "description": item.description,
                "minimum": item.minimum,
                "maximum": item.maximum,
                "default": 0,
            },
        )

        item = schema.IntegerItem(
            title="How many dogs",
            description="Question",
            minimum=0,
            maximum=10,
            default=0,
            enum=(0, 2, 4, 6),
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "integer",
                "title": item.title,
                "description": item.description,
                "minimum": item.minimum,
                "maximum": item.maximum,
                "default": 0,
                "enum": [0, 2, 4, 6],
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_integer_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.IntegerItem(title="How many dogs", description="Question")

        try:
            jsonschema.validate({"item": 2}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 3.1}, TestConf.serialize())
        self.assertIn("is not of type", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.IntegerItem(
                title="How many dogs", description="Question", multiple_of=2
            )

        try:
            jsonschema.validate({"item": 4}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 3}, TestConf.serialize())
        self.assertIn("is not a multiple of", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.IntegerItem(
                title="Foo", description="Foo Item", minimum=1, maximum=10
            )

        try:
            jsonschema.validate({"item": 3}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 11}, TestConf.serialize())
        self.assertIn("is greater than the maximum of", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.IntegerItem(
                title="Foo", description="Foo Item", minimum=10, maximum=100
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 3}, TestConf.serialize())
        self.assertIn("is less than the minimum of", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.IntegerItem(
                title="How many dogs",
                description="Question",
                minimum=0,
                exclusive_minimum=True,
                maximum=10,
                exclusive_maximum=True,
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 0}, TestConf.serialize())
        self.assertIn(
            "is less than or equal to the minimum of", excinfo.exception.message
        )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 10}, TestConf.serialize())
        self.assertIn(
            "is greater than or equal to the maximum of", excinfo.exception.message
        )

        class TestConf(schema.Schema):
            item = schema.IntegerItem(
                title="Foo", description="Foo Item", enum=(0, 2, 4, 6)
            )

        try:
            jsonschema.validate({"item": 4}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        class TestConf(schema.Schema):
            item = schema.IntegerItem(
                title="Foo", description="Foo Item", enum=(0, 2, 4, 6)
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 3}, TestConf.serialize())
        self.assertIn("is not one of", excinfo.exception.message)

    def test_array_config(self):
        string_item = schema.StringItem(title="Dog Name", description="The dog name")
        item = schema.ArrayItem(
            title="Dog Names", description="Name your dogs", items=string_item
        )

        self.assertDictEqual(
            item.serialize(),
            {
                "type": "array",
                "title": item.title,
                "description": item.description,
                "items": {
                    "type": "string",
                    "title": string_item.title,
                    "description": string_item.description,
                },
            },
        )

        integer_item = schema.IntegerItem(title="Dog Age", description="The dog age")
        item = schema.ArrayItem(
            title="Dog Names",
            description="Name your dogs",
            items=(string_item, integer_item),
        )

        self.assertDictEqual(
            item.serialize(),
            {
                "type": "array",
                "title": item.title,
                "description": item.description,
                "items": [
                    {
                        "type": "string",
                        "title": string_item.title,
                        "description": string_item.description,
                    },
                    {
                        "type": "integer",
                        "title": integer_item.title,
                        "description": integer_item.description,
                    },
                ],
            },
        )

        item = schema.ArrayItem(
            title="Dog Names",
            description="Name your dogs",
            items=(schema.StringItem(), schema.IntegerItem()),
            min_items=1,
            max_items=3,
            additional_items=False,
            unique_items=True,
        )

        self.assertDictEqual(
            item.serialize(),
            {
                "type": "array",
                "title": item.title,
                "description": item.description,
                "minItems": item.min_items,
                "maxItems": item.max_items,
                "uniqueItems": item.unique_items,
                "additionalItems": item.additional_items,
                "items": [{"type": "string"}, {"type": "integer"}],
            },
        )

        class HowManyConfig(schema.Schema):
            item = schema.IntegerItem(title="How many dogs", description="Question")

        item = schema.ArrayItem(
            title="Dog Names", description="Name your dogs", items=HowManyConfig()
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "array",
                "title": item.title,
                "description": item.description,
                "items": HowManyConfig.serialize(),
            },
        )

        class AgesConfig(schema.Schema):
            item = schema.IntegerItem()

        item = schema.ArrayItem(
            title="Dog Names",
            description="Name your dogs",
            items=(HowManyConfig(), AgesConfig()),
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "array",
                "title": item.title,
                "description": item.description,
                "items": [HowManyConfig.serialize(), AgesConfig.serialize()],
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_array_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.ArrayItem(
                title="Dog Names",
                description="Name your dogs",
                items=schema.StringItem(),
            )

        try:
            jsonschema.validate(
                {"item": ["Tobias", "Óscar"]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": ["Tobias", "Óscar", 3]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is not of type", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.ArrayItem(
                title="Dog Names",
                description="Name your dogs",
                items=schema.StringItem(),
                min_items=1,
                max_items=2,
            )

        try:
            jsonschema.validate(
                {"item": ["Tobias", "Óscar"]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": ["Tobias", "Óscar", "Pepe"]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is too long", excinfo.exception.message)

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": []},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is too short", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.ArrayItem(
                title="Dog Names",
                description="Name your dogs",
                items=schema.StringItem(),
                uniqueItems=True,
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": ["Tobias", "Tobias"]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("has non-unique elements", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.ArrayItem(items=(schema.StringItem(), schema.IntegerItem()))

        try:
            jsonschema.validate(
                {"item": ["Óscar", 4]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": ["Tobias", "Óscar"]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is not of type", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.ArrayItem(
                items=schema.ArrayItem(
                    items=(schema.StringItem(), schema.IntegerItem())
                )
            )

        try:
            jsonschema.validate(
                {"item": [["Tobias", 8], ["Óscar", 4]]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": [["Tobias", 8], ["Óscar", "4"]]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is not of type", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.ArrayItem(items=schema.StringItem(enum=["Tobias", "Óscar"]))

        try:
            jsonschema.validate(
                {"item": ["Óscar"]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")
        try:
            jsonschema.validate(
                {"item": ["Tobias"]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": ["Pepe"]},
                TestConf.serialize(),
                format_checker=jsonschema.FormatChecker(),
            )
        self.assertIn("is not one of", excinfo.exception.message)

    def test_dict_config(self):
        item = schema.DictItem(
            title="Poligon",
            description="Describe the Poligon",
            properties={"sides": schema.IntegerItem()},
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "object",
                "title": item.title,
                "description": item.description,
                "properties": {"sides": {"type": "integer"}},
            },
        )

        item = schema.DictItem(
            title="Poligon",
            description="Describe the Poligon",
            properties={"sides": schema.IntegerItem()},
            min_properties=1,
            max_properties=2,
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "object",
                "title": item.title,
                "description": item.description,
                "properties": {"sides": {"type": "integer"}},
                "minProperties": 1,
                "maxProperties": 2,
            },
        )

        item = schema.DictItem(
            title="Poligon",
            description="Describe the Poligon",
            pattern_properties={"s.*": schema.IntegerItem()},
            min_properties=1,
            max_properties=2,
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "object",
                "title": item.title,
                "description": item.description,
                "patternProperties": {"s.*": {"type": "integer"}},
                "minProperties": 1,
                "maxProperties": 2,
            },
        )

        item = schema.DictItem(
            title="Poligon",
            description="Describe the Poligon",
            properties={"color": schema.StringItem(enum=["red", "green", "blue"])},
            pattern_properties={"s*": schema.IntegerItem()},
            min_properties=1,
            max_properties=2,
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "object",
                "title": item.title,
                "description": item.description,
                "properties": {
                    "color": {"type": "string", "enum": ["red", "green", "blue"]}
                },
                "patternProperties": {"s*": {"type": "integer"}},
                "minProperties": 1,
                "maxProperties": 2,
            },
        )

        item = schema.DictItem(
            title="Poligon",
            description="Describe the Poligon",
            properties={"color": schema.StringItem(enum=["red", "green", "blue"])},
            pattern_properties={"s*": schema.IntegerItem()},
            additional_properties=True,
            min_properties=1,
            max_properties=2,
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "object",
                "title": item.title,
                "description": item.description,
                "properties": {
                    "color": {"type": "string", "enum": ["red", "green", "blue"]}
                },
                "patternProperties": {"s*": {"type": "integer"}},
                "minProperties": 1,
                "maxProperties": 2,
                "additionalProperties": True,
            },
        )
        item = schema.DictItem(
            title="Poligon",
            description="Describe the Poligon",
            properties={"sides": schema.IntegerItem()},
            additional_properties=schema.OneOfItem(
                items=[schema.BooleanItem(), schema.StringItem()]
            ),
        )
        self.assertDictEqual(
            item.serialize(),
            {
                "type": "object",
                "title": item.title,
                "description": item.description,
                "properties": {"sides": {"type": "integer"}},
                "additionalProperties": {
                    "oneOf": [{"type": "boolean"}, {"type": "string"}]
                },
            },
        )

        class TestConf(schema.Schema):
            item = schema.DictItem(
                title="Poligon",
                description="Describe the Poligon",
                properties={"sides": schema.IntegerItem(required=True)},
                additional_properties=schema.OneOfItem(
                    items=[schema.BooleanItem(), schema.StringItem()]
                ),
            )

        self.assertDictContainsSubset(
            TestConf.serialize(),
            {
                "$schema": "http://json-schema.org/draft-04/schema#",
                "type": "object",
                "properties": {
                    "item": {
                        "title": "Poligon",
                        "description": "Describe the Poligon",
                        "type": "object",
                        "properties": {"sides": {"type": "integer"}},
                        "additionalProperties": {
                            "oneOf": [{"type": "boolean"}, {"type": "string"}]
                        },
                        "required": ["sides"],
                    }
                },
                "x-ordering": ["item"],
                "additionalProperties": False,
            },
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_dict_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.DictItem(
                title="Poligon",
                description="Describe the Poligon",
                properties={"sides": schema.IntegerItem()},
            )

        try:
            jsonschema.validate({"item": {"sides": 1}}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": {"sides": "1"}}, TestConf.serialize())
        self.assertIn("is not of type", excinfo.exception.message)

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 2}, TestConf.serialize())
        self.assertIn("is not of type", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.DictItem(
                title="Poligon",
                description="Describe the Poligon",
                properties={"color": schema.StringItem(enum=["red", "green", "blue"])},
                pattern_properties={"si.*": schema.IntegerItem()},
            )

        try:
            jsonschema.validate(
                {"item": {"sides": 1, "color": "red"}}, TestConf.serialize()
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": {"sides": "4", "color": "blue"}}, TestConf.serialize()
            )
        self.assertIn("is not of type", excinfo.exception.message)

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 2}, TestConf.serialize())
        self.assertIn("is not of type", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.DictItem(
                title="Poligon",
                description="Describe the Poligon",
                properties={"color": schema.StringItem(enum=["red", "green", "blue"])},
                pattern_properties={"si.*": schema.IntegerItem()},
                additional_properties=False,
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": {"color": "green", "sides": 4, "surfaces": 4}},
                TestConf.serialize(),
            )
        if JSONSCHEMA_VERSION < Version("2.6.0"):
            self.assertIn(
                "Additional properties are not allowed", excinfo.exception.message
            )
        else:
            self.assertIn(
                "'surfaces' does not match any of the regexes",
                excinfo.exception.message,
            )

        class TestConf(schema.Schema):
            item = schema.DictItem(
                title="Poligon",
                description="Describe the Poligon",
                properties={"color": schema.StringItem(enum=["red", "green", "blue"])},
                additional_properties=schema.OneOfItem(
                    items=[schema.BooleanItem(), schema.IntegerItem()]
                ),
            )

        try:
            jsonschema.validate(
                {"item": {"sides": 1, "color": "red", "rugged_surface": False}},
                TestConf.serialize(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": {"sides": "4", "color": "blue"}}, TestConf.serialize()
            )
        if JSONSCHEMA_VERSION >= Version("3.0.0"):
            self.assertIn("'4'", excinfo.exception.message)
            self.assertIn("is not of type", excinfo.exception.message)
            self.assertIn("'boolean'", excinfo.exception.message)
        else:
            self.assertIn(
                "is not valid under any of the given schemas", excinfo.exception.message
            )

        class TestConf(schema.Schema):
            item = schema.DictItem(
                title="Poligon",
                description="Describe the Poligon",
                properties={"color": schema.StringItem(enum=["red", "green", "blue"])},
                additional_properties=schema.OneOfItem(
                    items=[schema.BooleanItem(), schema.IntegerItem()]
                ),
                min_properties=2,
                max_properties=3,
            )

        try:
            jsonschema.validate(
                {"item": {"color": "red", "sides": 1}}, TestConf.serialize()
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        try:
            jsonschema.validate(
                {"item": {"sides": 1, "color": "red", "rugged_surface": False}},
                TestConf.serialize(),
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": {"color": "blue"}}, TestConf.serialize())
        self.assertIn("does not have enough properties", excinfo.exception.message)

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {
                    "item": {
                        "sides": 4,
                        "color": "blue",
                        "rugged_surface": False,
                        "opaque": True,
                    }
                },
                TestConf.serialize(),
            )
        self.assertIn("has too many properties", excinfo.exception.message)

        class TestConf(schema.Schema):
            item = schema.DictItem(
                title="Poligon",
                description="Describe the Poligon",
                properties={"sides": schema.IntegerItem(required=True)},
                additional_properties=schema.OneOfItem(
                    items=[schema.BooleanItem(), schema.StringItem()]
                ),
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": {"color": "blue", "rugged_surface": False, "opaque": True}},
                TestConf.serialize(),
            )
        self.assertIn("'sides' is a required property", excinfo.exception.message)

        class Props(schema.Schema):
            sides = schema.IntegerItem(required=True)

        class TestConf(schema.Schema):
            item = schema.DictItem(
                title="Poligon",
                description="Describe the Poligon",
                properties=Props(),
                additional_properties=schema.OneOfItem(
                    items=[schema.BooleanItem(), schema.StringItem()]
                ),
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate(
                {"item": {"color": "blue", "rugged_surface": False, "opaque": True}},
                TestConf.serialize(),
            )
        self.assertIn("'sides' is a required property", excinfo.exception.message)

    def test_oneof_config(self):
        item = schema.OneOfItem(
            items=(
                schema.StringItem(title="Yes", enum=["yes"]),
                schema.StringItem(title="No", enum=["no"]),
            )
        )
        self.assertEqual(
            item.serialize(), {"oneOf": [i.serialize() for i in item.items]}
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_oneof_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.ArrayItem(
                title="Hungry",
                description="Are you hungry?",
                items=schema.OneOfItem(
                    items=(
                        schema.StringItem(title="Yes", enum=["yes"]),
                        schema.StringItem(title="No", enum=["no"]),
                    )
                ),
            )

        try:
            jsonschema.validate({"item": ["no"]}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": ["maybe"]}, TestConf.serialize())
        if JSONSCHEMA_VERSION >= Version("3.0.0"):
            self.assertIn("'maybe'", excinfo.exception.message)
            self.assertIn("is not one of", excinfo.exception.message)
            self.assertIn("'yes'", excinfo.exception.message)
        else:
            self.assertIn(
                "is not valid under any of the given schemas", excinfo.exception.message
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 2}, TestConf.serialize())
        self.assertIn("is not of type", excinfo.exception.message)

    def test_anyof_config(self):
        item = schema.AnyOfItem(
            items=(
                schema.StringItem(title="Yes", enum=["yes"]),
                schema.StringItem(title="No", enum=["no"]),
            )
        )
        self.assertEqual(
            item.serialize(),
            {"anyOf": [i.serialize() for i in item.items]},  # pylint: disable=E1133
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_anyof_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.ArrayItem(
                title="Hungry",
                description="Are you hungry?",
                items=schema.AnyOfItem(
                    items=(
                        schema.StringItem(title="Yes", enum=["yes"]),
                        schema.StringItem(title="No", enum=["no"]),
                        schema.BooleanItem(),
                    )
                ),
            )

        try:
            jsonschema.validate({"item": ["no"]}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        try:
            jsonschema.validate({"item": ["yes"]}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        try:
            jsonschema.validate({"item": [True]}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        try:
            jsonschema.validate({"item": [False]}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": ["maybe"]}, TestConf.serialize())
        if JSONSCHEMA_VERSION >= Version("3.0.0"):
            self.assertIn("'maybe'", excinfo.exception.message)
            self.assertIn("is not one of", excinfo.exception.message)
            self.assertIn("'yes'", excinfo.exception.message)
        else:
            self.assertIn(
                "is not valid under any of the given schemas", excinfo.exception.message
            )

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 2}, TestConf.serialize())
        self.assertIn("is not of type", excinfo.exception.message)

    def test_allof_config(self):
        item = schema.AllOfItem(
            items=(schema.StringItem(min_length=2), schema.StringItem(max_length=3))
        )
        self.assertEqual(
            item.serialize(),
            {"allOf": [i.serialize() for i in item.items]},  # pylint: disable=E1133
        )

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_allof_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.ArrayItem(
                title="Hungry",
                description="Are you hungry?",
                items=schema.AllOfItem(
                    items=(
                        schema.StringItem(min_length=2),
                        schema.StringItem(max_length=3),
                    )
                ),
            )

        try:
            jsonschema.validate({"item": ["no"]}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        try:
            jsonschema.validate({"item": ["yes"]}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": ["maybe"]}, TestConf.serialize())
        self.assertIn("is too long", excinfo.exception.message)

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": ["hmmmm"]}, TestConf.serialize())
        self.assertIn("is too long", excinfo.exception.message)

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": 2}, TestConf.serialize())
        self.assertIn("is not of type", excinfo.exception.message)

    def test_not_config(self):
        item = schema.NotItem(item=schema.BooleanItem())
        self.assertEqual(item.serialize(), {"not": item.item.serialize()})

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_not_config_validation(self):
        class TestConf(schema.Schema):
            item = schema.ArrayItem(
                title="Hungry",
                description="Are you hungry?",
                items=schema.NotItem(item=schema.BooleanItem()),
            )

        try:
            jsonschema.validate({"item": ["no"]}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        try:
            jsonschema.validate({"item": ["yes"]}, TestConf.serialize())
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": [True]}, TestConf.serialize())
        self.assertIn("is not allowed for", excinfo.exception.message)

        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"item": [False]}, TestConf.serialize())
        self.assertIn("is not allowed for", excinfo.exception.message)

    def test_item_name_override_class_attrname(self):
        class TestConf(schema.Schema):
            item = schema.BooleanItem(
                name="hungry", title="Hungry", description="Are you hungry?"
            )

        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "properties": {
                "hungry": {
                    "type": "boolean",
                    "description": "Are you hungry?",
                    "title": "Hungry",
                }
            },
            "x-ordering": ["hungry"],
            "additionalProperties": False,
        }
        self.assertDictEqual(TestConf.serialize(), expected)

    def test_config_name_override_class_attrname(self):
        class TestConf(schema.Schema):
            item = schema.BooleanItem(title="Hungry", description="Are you hungry?")

        class TestConf2(schema.Schema):
            a_name = TestConf(name="another_name")

        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "properties": {
                "another_name": {
                    "id": (
                        "https://non-existing.saltstack.com/schemas/another_name.json#"
                    ),
                    "type": "object",
                    "properties": {
                        "item": {
                            "type": "boolean",
                            "description": "Are you hungry?",
                            "title": "Hungry",
                        }
                    },
                    "x-ordering": ["item"],
                    "additionalProperties": False,
                }
            },
            "x-ordering": ["another_name"],
            "additionalProperties": False,
        }
        self.assertDictEqual(TestConf2.serialize(), expected)


class ComplexSchemaItem(schema.ComplexSchemaItem):
    _complex_attributes = ["thirsty"]
    thirsty = schema.BooleanItem(title="Thirsty", description="Are you thirsty?")


class ComplexComplexSchemaItem(schema.ComplexSchemaItem):
    _complex_attributes = ["hungry", "complex_item"]

    hungry = schema.BooleanItem(
        title="Hungry", description="Are you hungry?", required=True
    )
    complex_item = ComplexSchemaItem(definition_name="test_definition")


class TestComplexDefinitionsSchema(schema.DefinitionsSchema):
    title = "Test Complex Definition Schema"
    complex_item = ComplexSchemaItem()


class TestOneOfComplexDefinitionsSchema(schema.DefinitionsSchema):
    title = "Test OneOf Complex Definitions Schema"
    one_of_item = schema.OneOfItem(items=[ComplexSchemaItem(), schema.StringItem()])


class TestArrayComplexDefinitionsSchema(schema.DefinitionsSchema):
    title = "Test Array Complex Definitions Schema"
    array_item = schema.ArrayItem(items=ComplexSchemaItem())


class TestDictComplexDefinitionsSchema(schema.DefinitionsSchema):
    title = "Test Dict Complex Definitions Schema"
    dict_item = schema.DictItem(
        properties={"complex_obj": ComplexSchemaItem(required=True)},
        additional_properties=ComplexSchemaItem(),
    )


class TestComplexComplexDefinitionsSchema(schema.DefinitionsSchema):
    title = "Test Complex Complex Definition Schema"
    complex_complex_item = ComplexComplexSchemaItem()


class ComplexSchemaTestCase(TestCase):
    """Test cases with definition schemas containing complex items"""

    obj = ComplexSchemaItem()
    complex_obj = ComplexComplexSchemaItem()
    schema = TestComplexDefinitionsSchema()
    one_of_schema = TestOneOfComplexDefinitionsSchema()
    array_schema = TestArrayComplexDefinitionsSchema()
    dict_schema = TestDictComplexDefinitionsSchema()
    complex_schema = TestComplexComplexDefinitionsSchema()

    def test_complex_schema_item_serialize(self):
        obj = copy.deepcopy(self.obj)
        expected_serialized = {"$ref": "#/definitions/ComplexSchemaItem"}
        self.assertDictEqual(obj.serialize(), expected_serialized)

    def test_complex_schema_item_definition(self):
        obj = copy.deepcopy(self.obj)
        expected_def = {
            "type": "object",
            "title": "ComplexSchemaItem",
            "properties": {
                "thirsty": {
                    "type": "boolean",
                    "title": "Thirsty",
                    "description": "Are you thirsty?",
                }
            },
        }
        self.assertDictEqual(obj.get_definition(), expected_def)

    def test_complex_complex_schema_item_definition(self):
        complex_obj = copy.deepcopy(self.complex_obj)
        expected_def = {
            "type": "object",
            "title": "ComplexComplexSchemaItem",
            "properties": {
                "hungry": {
                    "type": "boolean",
                    "title": "Hungry",
                    "description": "Are you hungry?",
                },
                "complex_item": {
                    "type": "object",
                    "$ref": "#/definitions/test_definition",
                },
            },
            "required": ["hungry"],
        }
        self.assertDictEqual(complex_obj.get_definition(), expected_def)

    def test_complex_definition_schema(self):
        serialized = salt.utils.yaml.safe_load(
            salt.utils.json.dumps(self.schema.serialize())
        )
        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "Test Complex Definition Schema",
            "type": "object",
            "properties": {"complex_item": {"$ref": "#/definitions/ComplexSchemaItem"}},
            "x-ordering": ["complex_item"],
            "additionalProperties": False,
            "definitions": {
                "ComplexSchemaItem": {
                    "type": "object",
                    "title": "ComplexSchemaItem",
                    "properties": {
                        "thirsty": {
                            "type": "boolean",
                            "title": "Thirsty",
                            "description": "Are you thirsty?",
                        }
                    },
                }
            },
        }
        self.assertDictEqual(serialized, expected)

    def test_one_of_complex_definition_schema(self):
        serialized = salt.utils.yaml.safe_load(
            salt.utils.json.dumps(self.one_of_schema.serialize())
        )
        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "Test OneOf Complex Definitions Schema",
            "type": "object",
            "properties": {
                "one_of_item": {
                    "oneOf": [
                        {"$ref": "#/definitions/ComplexSchemaItem"},
                        {"type": "string"},
                    ]
                }
            },
            "x-ordering": ["one_of_item"],
            "additionalProperties": False,
            "definitions": {
                "ComplexSchemaItem": {
                    "type": "object",
                    "title": "ComplexSchemaItem",
                    "properties": {
                        "thirsty": {
                            "type": "boolean",
                            "title": "Thirsty",
                            "description": "Are you thirsty?",
                        }
                    },
                }
            },
        }
        self.assertDictEqual(serialized, expected)

    def test_array_complex_definition_schema(self):
        serialized = salt.utils.yaml.safe_load(
            salt.utils.json.dumps(self.array_schema.serialize())
        )
        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "Test Array Complex Definitions Schema",
            "type": "object",
            "properties": {
                "array_item": {
                    "type": "array",
                    "title": "array_item",
                    "items": {"$ref": "#/definitions/ComplexSchemaItem"},
                }
            },
            "x-ordering": ["array_item"],
            "additionalProperties": False,
            "definitions": {
                "ComplexSchemaItem": {
                    "type": "object",
                    "title": "ComplexSchemaItem",
                    "properties": {
                        "thirsty": {
                            "type": "boolean",
                            "title": "Thirsty",
                            "description": "Are you thirsty?",
                        }
                    },
                }
            },
        }
        self.assertDictEqual(serialized, expected)

    def test_dict_complex_definition_schema(self):
        serialized = salt.utils.yaml.safe_load(
            salt.utils.json.dumps(self.dict_schema.serialize())
        )
        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "Test Dict Complex Definitions Schema",
            "type": "object",
            "properties": {
                "dict_item": {
                    "type": "object",
                    "title": "dict_item",
                    "required": ["complex_obj"],
                    "properties": {
                        "complex_obj": {"$ref": "#/definitions/ComplexSchemaItem"}
                    },
                    "additionalProperties": {"$ref": "#/definitions/ComplexSchemaItem"},
                }
            },
            "x-ordering": ["dict_item"],
            "additionalProperties": False,
            "definitions": {
                "ComplexSchemaItem": {
                    "type": "object",
                    "title": "ComplexSchemaItem",
                    "properties": {
                        "thirsty": {
                            "type": "boolean",
                            "title": "Thirsty",
                            "description": "Are you thirsty?",
                        }
                    },
                }
            },
        }
        self.assertDictEqual(serialized, expected)

    def test_complex_complex_definition_schema(self):
        serialized = salt.utils.yaml.safe_load(
            salt.utils.json.dumps(self.complex_schema.serialize())
        )
        expected = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "Test Complex Complex Definition Schema",
            "type": "object",
            "properties": {
                "complex_complex_item": {
                    "$ref": "#/definitions/ComplexComplexSchemaItem"
                }
            },
            "x-ordering": ["complex_complex_item"],
            "additionalProperties": False,
            "definitions": {
                "ComplexComplexSchemaItem": {
                    "type": "object",
                    "title": "ComplexComplexSchemaItem",
                    "properties": {
                        "hungry": {
                            "type": "boolean",
                            "title": "Hungry",
                            "description": "Are you hungry?",
                        },
                        "complex_item": {
                            "type": "object",
                            "$ref": "#/definitions/test_definition",
                        },
                    },
                    "required": ["hungry"],
                },
                "test_definition": {
                    "type": "object",
                    "title": "test_definition",
                    "properties": {
                        "thirsty": {
                            "type": "boolean",
                            "title": "Thirsty",
                            "description": "Are you thirsty?",
                        }
                    },
                },
            },
        }
        self.assertDictEqual(serialized, expected)

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_complex_schema_item_thirsty_valid(self):
        serialized = self.schema.serialize()

        try:
            jsonschema.validate({"complex_item": {"thirsty": True}}, serialized)
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_complex_schema_item_thirsty_invalid(self):
        serialized = self.schema.serialize()
        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"complex_item": {"thirsty": "Foo"}}, serialized)
        expected = "'Foo' is not of type 'boolean'"
        self.assertIn(expected, excinfo.exception.message)

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_complex_complex_schema_item_hungry_valid(self):
        serialized = self.complex_schema.serialize()

        try:
            jsonschema.validate({"complex_complex_item": {"hungry": True}}, serialized)
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_both_complex_complex_schema_all_items_valid(self):
        serialized = self.complex_schema.serialize()
        try:
            jsonschema.validate(
                {
                    "complex_complex_item": {
                        "hungry": True,
                        "complex_item": {"thirsty": True},
                    }
                },
                serialized,
            )
        except jsonschema.exceptions.ValidationError as exc:
            self.fail(f"ValidationError raised: {exc}")

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_complex_complex_schema_item_hungry_invalid(self):
        serialized = self.complex_schema.serialize()
        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:
            jsonschema.validate({"complex_complex_item": {"hungry": "Foo"}}, serialized)
        expected = "'Foo' is not of type 'boolean'"
        self.assertIn(expected, excinfo.exception.message)

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_complex_complex_schema_item_inner_thirsty_invalid(self):
        serialized = self.complex_schema.serialize()
        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:

            jsonschema.validate(
                {
                    "complex_complex_item": {
                        "hungry": True,
                        "complex_item": {"thirsty": "Bar"},
                    }
                },
                serialized,
            )
        expected = "'Bar' is not of type 'boolean'"
        self.assertIn(expected, excinfo.exception.message)

    @pytest.mark.skipif(
        HAS_JSONSCHEMA is False, reason="The 'jsonschema' library is missing"
    )
    def test_complex_complex_schema_item_missing_required_hungry(self):
        serialized = self.complex_schema.serialize()
        with self.assertRaises(jsonschema.exceptions.ValidationError) as excinfo:

            jsonschema.validate(
                {"complex_complex_item": {"complex_item": {"thirsty": True}}},
                serialized,
            )
        self.assertIn("'hungry' is a required property", excinfo.exception.message)



import os

import pytest

from salt.runners import winrepo
from tests.support.mock import patch

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture
def configure_loader_modules(minion_opts, tmp_path):
    winrepo_dir = tmp_path / "winrepo"
    winrepo_dir.mkdir()
    winrepo_dir_ng = tmp_path / "winrepo_ng"
    winrepo_dir_ng.mkdir()
    minion_opts["winrepo_dir"] = str(winrepo_dir)
    minion_opts["winrepo_dir_ng"] = str(winrepo_dir_ng)
    return {winrepo: {"__opts__": minion_opts}}


@pytest.fixture
def winrepo_remotes(minion_opts):
    remotes = set()
    # Legacy repo has been archived as of September 2024
    remotes.update(minion_opts.get("winrepo_remotes", []))
    remotes.update(minion_opts.get("winrepo_remotes_ng", []))
    return remotes


def test_update_git_repos(winrepo_remotes):
    """
    Ensure update git repos works as intended.
    """
    res = winrepo.update_git_repos()
    assert res

    for remote in winrepo_remotes:
        assert remote in res
        assert res[remote]

        # Make sure there are package definitions in the root
        assert res[remote].endswith("_")
        pkg_def = os.path.join(res[remote], "7zip.sls")
        assert os.path.exists(pkg_def)


def test_legacy_update_git_repos(winrepo_remotes, minion_opts):
    """
    Ensure update git repos works as intended with legacy (non-gitfs) code.
    """
    with patch.object(winrepo, "_legacy_git", return_value=True):
        res = winrepo.update_git_repos()

        assert res

        for remote in winrepo_remotes:
            assert remote in res
            assert res[remote]

            # Make sure there are package definitions in the root
            # We have to look up the actual repo dir here because the legacy
            # update only returns True or False, not a path
            if "-ng" in remote:
                path = minion_opts["winrepo_dir_ng"]
                pkg_def = os.path.join(path, "salt-winrepo-ng", "_", "7zip.sls")
            else:
                path = minion_opts["winrepo_dir"]
                pkg_def = os.path.join(path, "salt-winrepo", "_", "7zip.sls")

            assert os.path.exists(pkg_def)



"""
Module for sending messages to Slack

.. versionadded:: 2015.5.0

:configuration: This module can be used by either passing an api key and version
    directly or by specifying both in a configuration profile in the salt
    master/minion config.

    For example:

    .. code-block:: yaml

        slack:
          api_key: peWcBiMOS9HrZG15peWcBiMOS9HrZG15
"""

import logging
import urllib.parse

import salt.utils.json
import salt.utils.slack
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)

__virtualname__ = "slack"


def __virtual__():
    """
    Return virtual name of the module.

    :return: The virtual name of the module.
    """
    return __virtualname__


def _get_api_key():
    api_key = __salt__["config.get"]("slack.api_key") or __salt__["config.get"](
        "slack:api_key"
    )

    if not api_key:
        raise SaltInvocationError("No Slack API key found.")

    return api_key


def _get_hook_id():
    url = __salt__["config.get"]("slack.hook") or __salt__["config.get"]("slack:hook")
    if not url:
        raise SaltInvocationError("No Slack WebHook url found")

    return url


def list_rooms(api_key=None):
    """
    List all Slack rooms.

    :param api_key: The Slack admin api key.
    :return: The room list.

    CLI Example:

    .. code-block:: bash

        salt '*' slack.list_rooms

        salt '*' slack.list_rooms api_key=peWcBiMOS9HrZG15peWcBiMOS9HrZG15
    """
    if not api_key:
        api_key = _get_api_key()
    return salt.utils.slack.query(function="rooms", api_key=api_key, opts=__opts__)


def list_users(api_key=None):
    """
    List all Slack users.

    :param api_key: The Slack admin api key.
    :return: The user list.

    CLI Example:

    .. code-block:: bash

        salt '*' slack.list_users

        salt '*' slack.list_users api_key=peWcBiMOS9HrZG15peWcBiMOS9HrZG15
    """
    if not api_key:
        api_key = _get_api_key()
    return salt.utils.slack.query(function="users", api_key=api_key, opts=__opts__)


def find_room(name, api_key=None):
    """
    Find a room by name and return it.

    :param name:    The room name.
    :param api_key: The Slack admin api key.
    :return:        The room object.

    CLI Example:

    .. code-block:: bash

        salt '*' slack.find_room name="random"

        salt '*' slack.find_room name="random" api_key=peWcBiMOS9HrZG15peWcBiMOS9HrZG15
    """
    if not api_key:
        api_key = _get_api_key()

    # search results don't include the name of the
    # channel with a hash, if the passed channel name
    # has a hash we remove it.
    if name.startswith("#"):
        name = name[1:]
    ret = list_rooms(api_key)
    if ret["res"]:
        rooms = ret["message"]
        if rooms:
            for room in rooms:
                if room["name"] == name:
                    return room
    return False


def find_user(name, api_key=None):
    """
    Find a user by name and return it.

    :param name:        The user name.
    :param api_key:     The Slack admin api key.
    :return:            The user object.

    CLI Example:

    .. code-block:: bash

        salt '*' slack.find_user name="ThomasHatch"

        salt '*' slack.find_user name="ThomasHatch" api_key=peWcBiMOS9HrZG15peWcBiMOS9HrZG15
    """
    if not api_key:
        api_key = _get_api_key()

    ret = list_users(api_key)
    if ret["res"]:
        users = ret["message"]
        if users:
            for user in users:
                if user["name"] == name:
                    return user
    return False


def post_message(
    channel,
    message,
    from_name,
    api_key=None,
    icon=None,
    attachments=None,
    blocks=None,
):
    """
    Send a message to a Slack channel.

    .. versionchanged:: 3003
        Added `attachments` and `blocks` kwargs

    :param channel:     The channel name, either will work.
    :param message:     The message to send to the Slack channel.
    :param from_name:   Specify who the message is from.
    :param api_key:     The Slack api key, if not specified in the configuration.
    :param icon:        URL to an image to use as the icon for this message
    :param attachments: Any attachments to be sent with the message.
    :param blocks:      Any blocks to be sent with the message.
    :return:            Boolean if message was sent successfully.

    CLI Example:

    .. code-block:: bash

        salt '*' slack.post_message channel="Development Room" message="Build is done" from_name="Build Server"

    """
    if not api_key:
        api_key = _get_api_key()

    if not channel:
        log.error("channel is a required option.")

    # channel must start with a hash or an @ (direct-message channels)
    if not channel.startswith("#") and not channel.startswith("@"):
        log.warning(
            "Channel name must start with a hash or @. "
            'Prepending a hash and using "#%s" as '
            "channel name instead of %s",
            channel,
            channel,
        )
        channel = f"#{channel}"

    if not from_name:
        log.error("from_name is a required option.")

    if not message:
        log.error("message is a required option.")

    if not from_name:
        log.error("from_name is a required option.")

    parameters = {
        "channel": channel,
        "username": from_name,
        "text": message,
        "attachments": attachments or [],
        "blocks": blocks or [],
    }

    if icon is not None:
        parameters["icon_url"] = icon

    # Slack wants the body on POST to be urlencoded.
    result = salt.utils.slack.query(
        function="message",
        api_key=api_key,
        method="POST",
        header_dict={"Content-Type": "application/x-www-form-urlencoded"},
        data=urllib.parse.urlencode(parameters),
        opts=__opts__,
    )

    if result["res"]:
        return True
    else:
        return result


def call_hook(
    message,
    attachment=None,
    color="good",
    short=False,
    identifier=None,
    channel=None,
    username=None,
    icon_emoji=None,
):
    """
    Send message to Slack incoming webhook.

    :param message:     The topic of message.
    :param attachment:  The message to send to the Slack WebHook.
    :param color:       The color of border of left side
    :param short:       An optional flag indicating whether the value is short
                        enough to be displayed side-by-side with other values.
    :param identifier:  The identifier of WebHook.
    :param channel:     The channel to use instead of the WebHook default.
    :param username:    Username to use instead of WebHook default.
    :param icon_emoji:  Icon to use instead of WebHook default.
    :return:            Boolean if message was sent successfully.

    CLI Example:

    .. code-block:: bash

        salt '*' slack.call_hook message='Hello, from SaltStack'

    """
    base_url = "https://hooks.slack.com/services/"
    if not identifier:
        identifier = _get_hook_id()

    url = urllib.parse.urljoin(base_url, identifier)

    if not message:
        log.error("message is required option")

    if attachment:
        payload = {
            "attachments": [
                {
                    "fallback": message,
                    "color": color,
                    "pretext": message,
                    "fields": [{"value": attachment, "short": short}],
                }
            ]
        }
    else:
        payload = {
            "text": message,
        }

    if channel:
        payload["channel"] = channel

    if username:
        payload["username"] = username

    if icon_emoji:
        payload["icon_emoji"] = icon_emoji

    data = urllib.parse.urlencode({"payload": salt.utils.json.dumps(payload)})
    result = salt.utils.http.query(url, method="POST", data=data, status=True)

    if result["status"] <= 201:
        return True
    else:
        return {"res": False, "message": result.get("body", result["status"])}



"""
:codeauthor: Shane Lee <slee@saltstack.com>
"""

import os

import pytest

import salt.modules.cmdmod as cmdmod
import salt.modules.win_file as win_file
import salt.modules.win_lgpo as win_lgpo
import salt.utils.win_dacl as win_dacl
import salt.utils.win_lgpo_auditpol as win_lgpo_auditpol
import salt.utils.win_reg as win_reg
from tests.support.mock import patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(tmp_path):
    cachedir = tmp_path / "__test_admx_policy_cache_dir"
    cachedir.mkdir(parents=True, exist_ok=True)
    return {
        win_lgpo: {
            "__salt__": {
                "cmd.run": cmdmod.run,
                "file.file_exists": win_file.file_exists,
                "file.makedirs": win_file.makedirs_,
                "file.remove": win_file.remove,
                "file.write": win_file.write,
            },
            "__opts__": {
                "cachedir": str(cachedir),
            },
            "__utils__": {
                "auditpol.get_auditpol_dump": win_lgpo_auditpol.get_auditpol_dump,
                "reg.read_value": win_reg.read_value,
            },
        },
        win_file: {
            "__utils__": {
                "dacl.set_perms": win_dacl.set_perms,
            },
        },
    }


def _test_mechanism(policy_name):
    """
    Helper function to get current setting
    """
    policy_data = win_lgpo._policy_info()
    policy_definition = policy_data.policies["Machine"]["policies"][policy_name]
    return win_lgpo._get_policy_info_setting(policy_definition)


def test_registry():
    """
    Test getting policy value using the Registry mechanism
    """
    policy_name = "RemoteRegistryExactPaths"
    result = _test_mechanism(policy_name=policy_name)
    expected = [
        "System\\CurrentControlSet\\Control\\ProductOptions",
        "System\\CurrentControlSet\\Control\\Server Applications",
        "Software\\Microsoft\\Windows NT\\CurrentVersion",
    ]
    assert result == expected


def test_secedit():
    """
    Test getting policy value using the Secedit mechanism
    """
    policy_name = "LSAAnonymousNameLookup"
    result = _test_mechanism(policy_name=policy_name)
    expected = "Disabled"
    assert result == expected


def test_netsh():
    """
    Test getting the policy value using the NetSH mechanism
    """
    policy_name = "WfwDomainState"
    all_settings = {
        "State": "NotConfigured",
        "Inbound": "NotConfigured",
        "Outbound": "NotConfigured",
        "LocalFirewallRules": "NotConfigured",
        "LocalConSecRules": "NotConfigured",
        "InboundUserNotification": "NotConfigured",
        "RemoteManagement": "NotConfigured",
        "UnicastResponseToMulticast": "NotConfigured",
        "LogAllowedConnections": "NotConfigured",
        "LogDroppedConnections": "NotConfigured",
        "FileName": "NotConfigured",
        "MaxFileSize": "NotConfigured",
    }
    with patch("salt.utils.win_lgpo_netsh.get_all_settings", return_value=all_settings):
        result = _test_mechanism(policy_name=policy_name)
    expected = "Not configured"
    assert result == expected


@pytest.mark.destructive_test
def test_adv_audit():
    """
    Test getting the policy value using the AdvAudit mechanism
    """
    system_root = os.environ.get("SystemRoot", "C:\\Windows")
    f_audit = os.path.join(system_root, "security", "audit", "audit.csv")
    f_audit_gpo = os.path.join(
        system_root,
        "System32",
        "GroupPolicy",
        "Machine",
        "Microsoft",
        "Windows NT",
        "Audit",
        "audit.csv",
    )
    if os.path.exists(f_audit):
        os.remove(f_audit)
    if os.path.exists(f_audit_gpo):
        os.remove(f_audit_gpo)
    policy_name = "AuditCredentialValidation"
    result = _test_mechanism(policy_name=policy_name)
    expected = "Not Configured"
    assert result == expected


def test_net_user_modal():
    """
    Test getting the policy value using the NetUserModal mechanism
    """
    policy_name = "PasswordHistory"
    result = _test_mechanism(policy_name=policy_name)
    expected = 0
    assert result == expected


def test_lsa_rights():
    """
    Test getting the policy value using the LsaRights mechanism
    """
    policy_name = "SeTrustedCredManAccessPrivilege"
    result = _test_mechanism(policy_name=policy_name)
    expected = []
    assert result == expected


def test_script_ini():
    """
    Test getting the policy value using the ScriptIni value
    """
    policy_name = "StartupScripts"
    result = _test_mechanism(policy_name=policy_name)
    expected = None
    assert result == expected



import salt.ext.tornado.concurrent
from salt.netapi.rest_tornado import saltnado


async def test_any_future():
    """
    Test that the Any Future does what we think it does
    """
    # create a few futures
    futures = []
    for _ in range(3):
        future = salt.ext.tornado.concurrent.Future()
        futures.append(future)

    # create an any future, make sure it isn't immediately done
    any_ = saltnado.Any(futures)
    assert any_.done() is False

    # finish one, lets see who finishes
    futures[0].set_result("foo")

    await futures[0]

    assert any_.done() is True
    assert futures[0].done() is True
    assert futures[1].done() is False
    assert futures[2].done() is False

    # make sure it returned the one that finished
    assert any_.result() == futures[0]

    futures = futures[1:]
    # re-wait on some other futures
    any_ = saltnado.Any(futures)
    futures[0].set_result("foo")
    await futures[0]

    assert any_.done() is True
    assert futures[0].done() is True
    assert futures[1].done() is False



"""
This module interacts with an ISC DHCP Server via OMAPI.
server_ip and server_port params may be set in the minion
config or pillar:

.. code-block:: yaml

  omapi.server_ip: 127.0.0.1
  omapi.server_port: 7991

:depends: pypureomapi Python module
"""

import logging
import struct

import salt.utils.stringutils

log = logging.getLogger(__name__)


try:
    import pypureomapi as omapi

    omapi_support = True
except ImportError as e:
    omapi_support = False


def __virtual__():
    """
    Confirm pypureomapi is available.
    """
    if omapi_support:
        return "omapi"
    return (
        False,
        "The omapi execution module cannot be loaded: "
        "the pypureomapi python library is not available.",
    )


def _conn():
    server_ip = __pillar__.get(
        "omapi.server_ip", __opts__.get("omapi.server_ip", "127.0.0.1")
    )
    server_port = __pillar__.get(
        "omapi.server_port", __opts__.get("omapi.server_port", 7991)
    )
    key = __pillar__.get("omapi.key", __opts__.get("omapi.key", None))
    username = __pillar__.get("omapi.user", __opts__.get("omapi.user", None))
    if key:
        key = salt.utils.stringutils.to_bytes(key)
    if username:
        username = salt.utils.stringutils.to_bytes(username)
    return omapi.Omapi(server_ip, server_port, username=username, key=key)


def add_host(mac, name=None, ip=None, ddns=False, group=None, supersede_host=False):
    """
    Add a host object for the given mac.

    CLI Example:

    .. code-block:: bash

        salt dhcp-server omapi.add_host ab:ab:ab:ab:ab:ab name=host1

    Add ddns-hostname and a fixed-ip statements:

    .. code-block:: bash

        salt dhcp-server omapi.add_host ab:ab:ab:ab:ab:ab name=host1 ip=10.1.1.1 ddns=true
    """
    statements = ""
    o = _conn()
    msg = omapi.OmapiMessage.open(b"host")
    msg.message.append((b"create", struct.pack(b"!I", 1)))
    msg.message.append((b"exclusive", struct.pack(b"!I", 1)))
    msg.obj.append((b"hardware-address", omapi.pack_mac(mac)))
    msg.obj.append((b"hardware-type", struct.pack(b"!I", 1)))
    if ip:
        msg.obj.append((b"ip-address", omapi.pack_ip(ip)))
    if name:
        msg.obj.append((b"name", salt.utils.stringutils.to_bytes(name)))
    if group:
        msg.obj.append((b"group", salt.utils.stringutils.to_bytes(group)))
    if supersede_host:
        statements += f'option host-name "{name}"; '
    if ddns and name:
        statements += f'ddns-hostname "{name}"; '
    if statements:
        msg.obj.append((b"statements", salt.utils.stringutils.to_bytes(statements)))
    response = o.query_server(msg)
    if response.opcode != omapi.OMAPI_OP_UPDATE:
        return False
    return True


def delete_host(mac=None, name=None):
    """
    Delete the host with the given mac or name.

    CLI Examples:

    .. code-block:: bash

        salt dhcp-server omapi.delete_host name=host1
        salt dhcp-server omapi.delete_host mac=ab:ab:ab:ab:ab:ab
    """
    if not (mac or name):
        raise TypeError("At least one argument is required")
    o = _conn()
    msg = omapi.OmapiMessage.open(b"host")
    if mac:
        msg.obj.append((b"hardware-address", omapi.pack_mac(mac)))
        msg.obj.append((b"hardware-type", struct.pack(b"!I", 1)))
    if name:
        msg.obj.append((b"name", salt.utils.stringutils.to_bytes(name)))
    response = o.query_server(msg)
    if response.opcode != omapi.OMAPI_OP_UPDATE:
        return None
    if response.handle == 0:
        return False
    response = o.query_server(omapi.OmapiMessage.delete(response.handle))
    if response.opcode != omapi.OMAPI_OP_STATUS:
        return False
    return True



import pytest

import salt.client.ssh.client
import salt.config
import salt.roster
import salt.utils.files
import salt.utils.path
import salt.utils.thin
import salt.utils.yaml
from tests.support.mock import MagicMock, patch


@pytest.fixture
def roster():
    return """
    localhost:
        host: 127.0.0.1
        port: 2827
    self:
        host: 0.0.0.0
        port: 42
    """


def test_roster_defaults_flat(tmp_path, roster):
    """
    Test Roster Defaults on the flat roster
    """
    expected = {
        "self": {"host": "0.0.0.0", "user": "daniel", "port": 42},
        "localhost": {"host": "127.0.0.1", "user": "daniel", "port": 2827},
    }

    root_dir = tmp_path / "foo" / "bar"
    root_dir.mkdir(exist_ok=True, parents=True)
    fpath = root_dir / "config"
    with salt.utils.files.fopen(str(fpath), "w") as fp_:
        fp_.write(
            """
            roster_defaults:
                user: daniel
            """
        )
    opts = salt.config.master_config(fpath)
    with patch("salt.roster.get_roster_file", MagicMock(return_value=roster)):
        with patch(
            "salt.template.compile_template",
            MagicMock(return_value=salt.utils.yaml.safe_load(roster)),
        ):
            roster = salt.roster.Roster(opts=opts)
            assert roster.targets("*", "glob") == expected



#
# Author: Bo Maryniuk <bo@suse.de>
#
"""
Ansible Support
===============

This module can have an optional minion-level
configuration in /etc/salt/minion.d/ as follows:

  ansible_timeout: 1200

The timeout is how many seconds Salt should wait for
any Ansible module to respond.
"""

import fnmatch
import json
import logging
import os
import subprocess
import sys
from tempfile import NamedTemporaryFile

import salt.utils.ansible
import salt.utils.decorators.path
import salt.utils.json
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.timed_subprocess
import salt.utils.yaml
from salt.exceptions import CommandExecutionError

# Function alias to make sure not to shadow built-in's
__func_alias__ = {"list_": "list"}

__virtualname__ = "ansible"

log = logging.getLogger(__name__)

INVENTORY = """
hosts:
   vars:
     ansible_connection: local
"""
DEFAULT_TIMEOUT = 1200  # seconds (20 minutes)

__non_ansible_functions__ = []

__load__ = __non_ansible_functions__[:] = [
    "help",
    "list_",
    "call",
    "playbooks",
    "discover_playbooks",
    "targets",
]


def _set_callables(modules):
    """
    Set all Ansible modules callables
    :return:
    """

    def _set_function(real_cmd_name, doc):
        """
        Create a Salt function for the Ansible module.
        """

        def _cmd(*args, **kwargs):
            """
            Call an Ansible module as a function from the Salt.
            """
            return call(real_cmd_name, *args, **kwargs)

        _cmd.__doc__ = doc
        return _cmd

    for mod, (real_mod, doc) in modules.items():
        __load__.append(mod)
        setattr(sys.modules[__name__], mod, _set_function(real_mod, doc))


def __virtual__():
    if salt.utils.platform.is_windows():
        return False, "The ansiblegate module isn't supported on Windows"
    ansible_bin = salt.utils.path.which("ansible")
    if not ansible_bin:
        return False, "The 'ansible' binary was not found."
    ansible_doc_bin = salt.utils.path.which("ansible-doc")
    if not ansible_doc_bin:
        return False, "The 'ansible-doc' binary was not found."
    ansible_playbook_bin = salt.utils.path.which("ansible-playbook")
    if not ansible_playbook_bin:
        return False, "The 'ansible-playbook' binary was not found."

    env = os.environ.copy()
    env["ANSIBLE_DEPRECATION_WARNINGS"] = "0"

    proc = subprocess.run(
        [ansible_doc_bin, "--list", "--json", "--type=module"],
        capture_output=True,
        check=False,
        shell=False,
        text=True,
        env=env,
    )
    if proc.returncode != 0:
        return (
            False,
            f"Failed to get the listing of ansible modules:\n{proc.stderr}",
        )

    module_funcs = dir(sys.modules[__name__])
    ansible_module_listing = salt.utils.json.loads(proc.stdout)
    salt_ansible_modules_mapping = {}
    for key in list(ansible_module_listing):
        if not key.startswith("ansible."):
            salt_ansible_modules_mapping[key] = (key, ansible_module_listing[key])
            continue

        # Strip 'ansible.' from the module
        # Fyi, str.partition() is faster than str.replace()
        _, _, alias = key.partition(".")
        if alias in salt_ansible_modules_mapping:
            continue
        if alias in module_funcs:
            continue
        salt_ansible_modules_mapping[alias] = (key, ansible_module_listing[key])
        if alias.startswith(("builtin.", "system.")):
            # Strip "builtin." or "system." so that we can do something like
            # "salt-call ansible.ping" instead of "salt-call ansible.builtin.ping",
            # although both formats can be used
            _, _, alias = alias.partition(".")
            if alias in salt_ansible_modules_mapping:
                continue
            if alias in module_funcs:
                continue
            salt_ansible_modules_mapping[alias] = (key, ansible_module_listing[key])

    _set_callables(salt_ansible_modules_mapping)
    return __virtualname__


def help(module=None, *args):
    """
    Display help on Ansible standard module.

    :param module: The module to get the help

    CLI Example:

    .. code-block:: bash

        salt * ansible.help ping
    """
    if not module:
        raise CommandExecutionError(
            "Please tell me what module you want to have helped with. "
            'Or call "ansible.list" to know what is available.'
        )

    ansible_doc_bin = salt.utils.path.which("ansible-doc")

    env = os.environ.copy()
    env["ANSIBLE_DEPRECATION_WARNINGS"] = "0"

    proc = subprocess.run(
        [ansible_doc_bin, "--json", "--type=module", module],
        capture_output=True,
        check=True,
        shell=False,
        text=True,
        env=env,
    )
    data = salt.utils.json.loads(proc.stdout)
    doc = data[next(iter(data))]
    if not args:
        ret = doc["doc"]
        for section in ("examples", "return", "metadata"):
            section_data = doc.get(section)
            if section_data:
                ret[section] = section_data
    else:
        ret = {}
        for arg in args:
            info = doc.get(arg)
            if info is not None:
                ret[arg] = info
    return ret


def list_(pattern=None):
    """
    Lists available modules.

    CLI Example:

    .. code-block:: bash

        salt * ansible.list
        salt * ansible.list '*win*'  # To get all modules matching 'win' on it's name
    """
    if pattern is None:
        module_list = set(__load__)
        module_list.discard(set(__non_ansible_functions__))
        return sorted(module_list)
    return sorted(fnmatch.filter(__load__, pattern))


def call(module, *args, **kwargs):
    """
    Call an Ansible module by invoking it.

    :param module: the name of the module.
    :param args: Arguments to pass to the module
    :param kwargs: keywords to pass to the module

    CLI Example:

    .. code-block:: bash

        salt * ansible.call ping data=foobar
    """

    module_args = []
    for arg in args:
        module_args.append(salt.utils.json.dumps(arg))

    _kwargs = {}
    for _kw in kwargs.get("__pub_arg", []):
        if isinstance(_kw, dict):
            _kwargs = _kw
            break
    else:
        _kwargs = {k: v for (k, v) in kwargs.items() if not k.startswith("__pub")}

    for key, value in _kwargs.items():
        module_args.append(f"{key}={salt.utils.json.dumps(value)}")

    with NamedTemporaryFile(mode="w") as inventory:

        ansible_binary_path = salt.utils.path.which("ansible")
        log.debug("Calling ansible module %r", module)
        try:
            env = os.environ.copy()
            env["ANSIBLE_DEPRECATION_WARNINGS"] = "0"

            proc_exc = subprocess.run(
                [
                    ansible_binary_path,
                    "localhost",
                    "--limit",
                    "127.0.0.1",
                    "-m",
                    module,
                    "-a",
                    " ".join(module_args),
                    "-i",
                    inventory.name,
                ],
                capture_output=True,
                timeout=__opts__.get("ansible_timeout", DEFAULT_TIMEOUT),
                text=True,
                check=True,
                shell=False,
                env=env,
            )

            original_output = proc_exc.stdout
            proc_out = original_output.splitlines()
            if proc_out[0].endswith("{"):
                proc_out[0] = "{"
                try:
                    out = salt.utils.json.loads("\n".join(proc_out))
                except ValueError as exc:
                    out = {
                        "Error": proc_exc.stderr or str(exc),
                        "Output": original_output,
                    }
                    return out
            elif proc_out[0].endswith(">>"):
                out = {"output": "\n".join(proc_out[1:])}
            else:
                out = {"output": original_output}

        except subprocess.CalledProcessError as exc:
            out = {"Exitcode": exc.returncode, "Error": exc.stderr or str(exc)}
            if exc.stdout:
                out["Given JSON output"] = exc.stdout
            return out

    for key in ("invocation", "changed"):
        out.pop(key, None)

    return out


@salt.utils.decorators.path.which("ansible-playbook")
def playbooks(
    playbook,
    rundir=None,
    check=False,
    diff=False,
    extra_vars=None,
    flush_cache=False,
    forks=5,
    inventory=None,
    limit=None,
    list_hosts=False,
    list_tags=False,
    list_tasks=False,
    module_path=None,
    skip_tags=None,
    start_at_task=None,
    syntax_check=False,
    tags=None,
    playbook_kwargs=None,
):
    """
    Run Ansible Playbooks

    :param playbook: Which playbook to run.
    :param rundir: Directory to run `ansible-playbook` in. (Default: None)
    :param check: don't make any changes; instead, try to predict some
                  of the changes that may occur (Default: False)
    :param diff: when changing (small) files and templates, show the
                 differences in those files; works great with --check
                 (default: False)
    :param extra_vars: set additional variables as key=value or YAML/JSON, if
                       filename prepend with @, (default: None)
    :param flush_cache: clear the fact cache for every host in inventory
                        (default: False)
    :param forks: specify number of parallel processes to use
                  (Default: 5)
    :param inventory: specify inventory host path or comma separated host
                      list. (Default: None) (Ansible's default is /etc/ansible/hosts)
    :param limit: further limit selected hosts to an additional pattern (Default: None)
    :param list_hosts: outputs a list of matching hosts; does not execute anything else
                       (Default: False)
    :param list_tags: list all available tags (Default: False)
    :param list_tasks: list all tasks that would be executed (Default: False)
    :param module_path: prepend colon-separated path(s) to module library. (Default: None)
    :param skip_tags: only run plays and tasks whose tags do not match these
                      values (Default: False)
    :param start_at_task: start the playbook at the task matching this name (Default: None)
    :param: syntax_check: perform a syntax check on the playbook, but do not execute it
                          (Default: False)
    :param tags: only run plays and tasks tagged with these values (Default: None)

    :return: Playbook return

    CLI Example:

    .. code-block:: bash

        salt 'ansiblehost'  ansible.playbooks playbook=/srv/playbooks/play.yml
    """
    command = ["ansible-playbook", playbook]
    if check:
        command.append("--check")
    if diff:
        command.append("--diff")
    if isinstance(extra_vars, dict):
        command.append(f"--extra-vars='{json.dumps(extra_vars)}'")
    elif isinstance(extra_vars, str) and extra_vars.startswith("@"):
        command.append(f"--extra-vars={extra_vars}")
    if flush_cache:
        command.append("--flush-cache")
    if inventory:
        command.append(f"--inventory={inventory}")
    if limit:
        command.append(f"--limit={limit}")
    if list_hosts:
        command.append("--list-hosts")
    if list_tags:
        command.append("--list-tags")
    if list_tasks:
        command.append("--list-tasks")
    if module_path:
        command.append(f"--module-path={module_path}")
    if skip_tags:
        command.append(f"--skip-tags={skip_tags}")
    if start_at_task:
        command.append(f"--start-at-task={start_at_task}")
    if syntax_check:
        command.append("--syntax-check")
    if tags:
        command.append(f"--tags={tags}")
    if playbook_kwargs:
        for key, value in playbook_kwargs.items():
            key = key.replace("_", "-")
            if value is True:
                command.append(f"--{key}")
            elif isinstance(value, str):
                command.append(f"--{key}={value}")
            elif isinstance(value, dict):
                command.append(f"--{key}={json.dumps(value)}")
    command.append(f"--forks={forks}")
    cmd_kwargs = {
        "env": {
            "ANSIBLE_STDOUT_CALLBACK": "json",
            "ANSIBLE_RETRY_FILES_ENABLED": "0",
            "ANSIBLE_DEPRECATION_WARNINGS": "0",
        },
        "cwd": rundir,
        "cmd": " ".join(command),
        "reset_system_locale": False,
    }
    ret = __salt__["cmd.run_all"](**cmd_kwargs)
    log.debug("Ansible Playbook Return: %s", ret)
    try:
        retdata = json.loads(ret["stdout"])
    except ValueError:
        retdata = ret
    if "retcode" in ret:
        __context__["retcode"] = retdata["retcode"] = ret["retcode"]
    return retdata


def targets(inventory="/etc/ansible/hosts", yaml=False, export=False):
    """
    .. versionadded:: 3005

    Return the inventory from an Ansible inventory_file

    :param inventory:
        The inventory file to read the inventory from. Default: "/etc/ansible/hosts"

    :param yaml:
        Return the inventory as yaml output. Default: False

    :param export:
        Return inventory as export format. Default: False

    CLI Example:

    .. code-block:: bash

        salt 'ansiblehost' ansible.targets
        salt 'ansiblehost' ansible.targets inventory=my_custom_inventory

    """
    return salt.utils.ansible.targets(inventory=inventory, yaml=yaml, export=export)


def discover_playbooks(
    path=None,
    locations=None,
    playbook_extension=None,
    hosts_filename=None,
    syntax_check=False,
):
    """
    .. versionadded:: 3005

    Discover Ansible playbooks stored under the given path or from multiple paths (locations)

    This will search for files matching with the playbook file extension under the given
    root path and will also look for files inside the first level of directories in this path.

    The return of this function would be a dict like this:

    .. code-block:: python

        {
            "/home/foobar/": {
                "my_ansible_playbook.yml": {
                    "fullpath": "/home/foobar/playbooks/my_ansible_playbook.yml",
                    "custom_inventory": "/home/foobar/playbooks/hosts"
                },
                "another_playbook.yml": {
                    "fullpath": "/home/foobar/playbooks/another_playbook.yml",
                    "custom_inventory": "/home/foobar/playbooks/hosts"
                },
                "lamp_simple/site.yml": {
                    "fullpath": "/home/foobar/playbooks/lamp_simple/site.yml",
                    "custom_inventory": "/home/foobar/playbooks/lamp_simple/hosts"
                },
                "lamp_proxy/site.yml": {
                    "fullpath": "/home/foobar/playbooks/lamp_proxy/site.yml",
                    "custom_inventory": "/home/foobar/playbooks/lamp_proxy/hosts"
                }
            },
            "/srv/playbooks/": {
                "example_playbook/example.yml": {
                    "fullpath": "/srv/playbooks/example_playbook/example.yml",
                    "custom_inventory": "/srv/playbooks/example_playbook/hosts"
                }
            }
        }

    :param path:
        Path to discover playbooks from.

    :param locations:
        List of paths to discover playbooks from.

    :param playbook_extension:
        File extension of playbooks file to search for. Default: "yml"

    :param hosts_filename:
        Filename of custom playbook inventory to search for. Default: "hosts"

    :param syntax_check:
        Skip playbooks that do not pass "ansible-playbook --syntax-check" validation. Default: False

    :return:
        The discovered playbooks under the given paths

    CLI Example:

    .. code-block:: bash

        salt 'ansiblehost' ansible.discover_playbooks path=/srv/playbooks/
        salt 'ansiblehost' ansible.discover_playbooks locations='["/srv/playbooks/", "/srv/foobar"]'

    """

    if not path and not locations:
        raise CommandExecutionError(
            "You have to specify either 'path' or 'locations' arguments"
        )

    if path and locations:
        raise CommandExecutionError(
            "You cannot specify 'path' and 'locations' at the same time"
        )

    if not playbook_extension:
        playbook_extension = "yml"
    if not hosts_filename:
        hosts_filename = "hosts"

    if path:
        if not os.path.isabs(path):
            raise CommandExecutionError(
                f"The given path is not an absolute path: {path}"
            )
        if not os.path.isdir(path):
            raise CommandExecutionError(f"The given path is not a directory: {path}")
        return {
            path: _explore_path(path, playbook_extension, hosts_filename, syntax_check)
        }

    if locations:
        all_ret = {}
        for location in locations:
            all_ret[location] = _explore_path(
                location, playbook_extension, hosts_filename, syntax_check
            )
        return all_ret


def _explore_path(path, playbook_extension, hosts_filename, syntax_check):
    ret = {}

    if not os.path.isabs(path):
        log.error("The given path is not an absolute path: %s", path)
        return ret
    if not os.path.isdir(path):
        log.error("The given path is not a directory: %s", path)
        return ret

    try:
        # Check files in the given path
        for _f in os.listdir(path):
            _path = os.path.join(path, _f)
            if os.path.isfile(_path) and _path.endswith("." + playbook_extension):
                ret[_f] = {"fullpath": _path}
                # Check for custom inventory file
                if os.path.isfile(os.path.join(path, hosts_filename)):
                    ret[_f].update(
                        {"custom_inventory": os.path.join(path, hosts_filename)}
                    )
            elif os.path.isdir(_path):
                # Check files in the 1st level of subdirectories
                for _f2 in os.listdir(_path):
                    _path2 = os.path.join(_path, _f2)
                    if os.path.isfile(_path2) and _path2.endswith(
                        "." + playbook_extension
                    ):
                        ret[os.path.join(_f, _f2)] = {"fullpath": _path2}
                        # Check for custom inventory file
                        if os.path.isfile(os.path.join(_path, hosts_filename)):
                            ret[os.path.join(_f, _f2)].update(
                                {
                                    "custom_inventory": os.path.join(
                                        _path, hosts_filename
                                    )
                                }
                            )
    except Exception as exc:
        raise CommandExecutionError(
            f"There was an exception while discovering playbooks: {exc}"
        )

    # Run syntax check validation
    if syntax_check:
        check_command = ["ansible-playbook", "--syntax-check"]
        try:
            for pb in list(ret):
                if __salt__["cmd.retcode"](
                    check_command + [ret[pb]], reset_system_locale=False
                ):
                    del ret[pb]
        except Exception as exc:
            raise CommandExecutionError(
                "There was an exception while checking syntax of playbooks: {}".format(
                    exc
                )
            )
    return ret



"""
    :codeauthor: Mike Place (mp@saltstack.com)

    Test SMTP returner
"""

import pytest

import salt.returners.smtp_return as smtp
from salt.utils.jinja import SaltCacheLoader
from tests.support.mock import MagicMock, patch

try:
    import gnupg  # pylint: disable=unused-import

    HAS_GNUPG = True
except ImportError:
    HAS_GNUPG = False


@pytest.fixture
def configure_loader_modules():
    return {
        smtp: {
            "__opts__": {
                "extension_modules": "",
                "optimization_order": [0, 1, 2],
                "renderer": "jinja|yaml",
                "renderer_blacklist": [],
                "renderer_whitelist": [],
                "file_roots": {},
                "pillar_roots": {},
                "cachedir": "/",
                "master_uri": "tcp://127.0.0.1:4505",
                "pki_dir": "/",
                "keysize": 2048,
                "id": "test",
                "__role": "minion",
            }
        }
    }


def _test_returner(mocked_smtplib):  # pylint: disable=unused-argument
    """
    Test to see if the SMTP returner sends a message
    """
    ret = {
        "id": "12345",
        "fun": "mytest.func",
        "fun_args": "myfunc args",
        "jid": "54321",
        "return": "The room is on fire as shes fixing her hair",
    }
    options = {
        "username": "",
        "tls": "",
        "from": "",
        "fields": "id,fun,fun_args,jid,return",
        "to": "",
        "host": "",
        "renderer": "jinja|yaml",
        "template": "",
        "password": "",
        "gpgowner": "",
        "subject": "",
    }

    with patch(
        "salt.returners.smtp_return._get_options", MagicMock(return_value=options)
    ), patch.object(SaltCacheLoader, "file_client", MagicMock()):
        smtp.returner(ret)
        assert mocked_smtplib.return_value.sendmail.called is True


@pytest.mark.skipif(not HAS_GNUPG, reason="Need gnupg to run this test")
def test_returner_gnupg():
    with patch("salt.returners.smtp_return.gnupg"), patch(
        "salt.returners.smtp_return.smtplib.SMTP"
    ) as mocked_smtplib:
        _test_returner(mocked_smtplib)


@pytest.mark.skipif(HAS_GNUPG, reason="Only run this test without gnupg")
def test_returner_no_gnupg():
    with patch("salt.returners.smtp_return.smtplib.SMTP") as mocked_smtplib:
        _test_returner(mocked_smtplib)



import os

import pytest

import salt.utils.cache
import salt.utils.files
import salt.utils.path
import salt.version

_DUMMY_FILES = (
    "data.txt",
    "foo.t2",
    "bar.t3",
    "nested/test",
    "nested/cache.txt",
    "n/n1/n2/n3/n4/n5",
)


def _make_dummy_files(tmp_path):
    for full_path in _DUMMY_FILES:
        full_path = salt.utils.path.join(tmp_path, full_path)
        path, _ = os.path.split(full_path)
        if not os.path.isdir(path):
            os.makedirs(path)
        with salt.utils.files.fopen(full_path, "w") as file:
            file.write("data")


def _dummy_files_exists(tmp_path):
    """
    True if all files exists
    False if all files are missing
    None if some files exists and others are missing
    """
    ret = None
    for full_path in _DUMMY_FILES:
        full_path = salt.utils.path.join(tmp_path, full_path)
        is_file = os.path.isfile(full_path)
        if ret is None:
            ret = is_file
        elif ret is not is_file:
            return None  # Some files are found and others are missing
    return ret


def test_verify_cache_version_bad_path():
    with pytest.raises(ValueError):
        # cache version should fail if given bad file python
        salt.utils.cache.verify_cache_version("\0/bad/path")


def test_verify_cache_version(tmp_path):
    # cache version should make dir if it does not exist
    tmp_path = str(salt.utils.path.join(str(tmp_path), "work", "salt"))
    cache_version = salt.utils.path.join(tmp_path, "cache_version")

    # check that cache clears when no cache_version is present
    _make_dummy_files(tmp_path)
    assert salt.utils.cache.verify_cache_version(tmp_path) is False
    assert _dummy_files_exists(tmp_path) is False

    # check that cache_version has correct salt version
    with salt.utils.files.fopen(cache_version, "r") as file:
        assert "\n".join(file.readlines()) == salt.version.__version__

    # check that cache does not get clear when check is called multiple times
    _make_dummy_files(tmp_path)
    for _ in range(3):
        assert salt.utils.cache.verify_cache_version(tmp_path) is True
        assert _dummy_files_exists(tmp_path) is True

    # check that cache clears when a different version is present
    with salt.utils.files.fopen(cache_version, "w") as file:
        file.write("-1")
    assert salt.utils.cache.verify_cache_version(tmp_path) is False
    assert _dummy_files_exists(tmp_path) is False

    # check that cache does not get clear when check is called multiple times
    _make_dummy_files(tmp_path)
    for _ in range(3):
        assert salt.utils.cache.verify_cache_version(tmp_path) is True
        assert _dummy_files_exists(tmp_path) is True



"""
Connection module for Amazon EFS

.. versionadded:: 2017.7.0

:configuration: This module accepts explicit EFS credentials but can also
    utilize IAM roles assigned to the instance through Instance Profiles or
    it can read them from the ~/.aws/credentials file or from these
    environment variables: AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary.  More information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/efs/latest/ug/
            access-control-managing-permissions.html

        http://boto3.readthedocs.io/en/latest/guide/
            configuration.html#guide-configuration

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file

    .. code-block:: yaml

        efs.keyid: GKTADJGHEIQSXMKKRBJ08H
        efs.key: askd+ghsdfjkghWupU/asdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration

    .. code-block:: yaml

        efs.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid, and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
          keyid: GKTADJGHEIQSXMKKRBJ08H
          key: askd+ghsdfjkghWupU/asdflkdfklgjsdfjajkghs
          region: us-east-1

:depends: boto3
"""

import logging

import salt.utils.versions

try:
    import boto3

    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False


log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto3 libraries exist and if boto3 libraries are greater than
    a given version.
    """
    return salt.utils.versions.check_boto_reqs(boto3_ver="1.0.0", check_boto=False)


def _get_conn(key=None, keyid=None, profile=None, region=None, **kwargs):
    """
    Create a boto3 client connection to EFS
    """
    client = None
    if profile:
        if isinstance(profile, str):
            if profile in __pillar__:
                profile = __pillar__[profile]
            elif profile in __opts__:
                profile = __opts__[profile]
    elif key or keyid or region:
        profile = {}
        if key:
            profile["key"] = key
        if keyid:
            profile["keyid"] = keyid
        if region:
            profile["region"] = region

    if isinstance(profile, dict):
        if "region" in profile:
            profile["region_name"] = profile["region"]
            profile.pop("region", None)
        if "key" in profile:
            profile["aws_secret_access_key"] = profile["key"]
            profile.pop("key", None)
        if "keyid" in profile:
            profile["aws_access_key_id"] = profile["keyid"]
            profile.pop("keyid", None)

        client = boto3.client("efs", **profile)
    else:
        client = boto3.client("efs")

    return client


def create_file_system(
    name,
    performance_mode="generalPurpose",
    keyid=None,
    key=None,
    profile=None,
    region=None,
    creation_token=None,
    **kwargs
):
    """
    Creates a new, empty file system.

    name
        (string) - The name for the new file system

    performance_mode
        (string) - The PerformanceMode of the file system. Can be either
        generalPurpose or maxIO

    creation_token
        (string) - A unique name to be used as reference when creating an EFS.
        This will ensure idempotency. Set to name if not specified otherwise

    returns
        (dict) - A dict of the data for the elastic file system

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' boto_efs.create_file_system efs-name generalPurpose
    """

    if creation_token is None:
        creation_token = name

    tags = {"Key": "Name", "Value": name}

    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)

    response = client.create_file_system(
        CreationToken=creation_token, PerformanceMode=performance_mode
    )

    if "FileSystemId" in response:
        client.create_tags(FileSystemId=response["FileSystemId"], Tags=tags)

    if "Name" in response:
        response["Name"] = name

    return response


def create_mount_target(
    filesystemid,
    subnetid,
    ipaddress=None,
    securitygroups=None,
    keyid=None,
    key=None,
    profile=None,
    region=None,
    **kwargs
):
    """
    Creates a mount target for a file system.
    You can then mount the file system on EC2 instances via the mount target.

    You can create one mount target in each Availability Zone in your VPC.
    All EC2 instances in a VPC within a given Availability Zone share a
    single mount target for a given file system.

    If you have multiple subnets in an Availability Zone,
    you create a mount target in one of the subnets.
    EC2 instances do not need to be in the same subnet as the mount target
    in order to access their file system.

    filesystemid
        (string) - ID of the file system for which to create the mount target.

    subnetid
        (string) - ID of the subnet to add the mount target in.

    ipaddress
        (string) - Valid IPv4 address within the address range
                    of the specified subnet.

    securitygroups
        (list[string]) - Up to five VPC security group IDs,
                            of the form sg-xxxxxxxx.
                            These must be for the same VPC as subnet specified.

    returns
        (dict) - A dict of the response data

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' boto_efs.create_mount_target filesystemid subnetid
    """

    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)

    if ipaddress is None and securitygroups is None:
        return client.create_mount_target(FileSystemId=filesystemid, SubnetId=subnetid)

    if ipaddress is None:
        return client.create_mount_target(
            FileSystemId=filesystemid, SubnetId=subnetid, SecurityGroups=securitygroups
        )
    if securitygroups is None:
        return client.create_mount_target(
            FileSystemId=filesystemid, SubnetId=subnetid, IpAddress=ipaddress
        )

    return client.create_mount_target(
        FileSystemId=filesystemid,
        SubnetId=subnetid,
        IpAddress=ipaddress,
        SecurityGroups=securitygroups,
    )


def create_tags(
    filesystemid, tags, keyid=None, key=None, profile=None, region=None, **kwargs
):
    """
    Creates or overwrites tags associated with a file system.
    Each tag is a key-value pair. If a tag key specified in the request
    already exists on the file system, this operation overwrites
    its value with the value provided in the request.

    filesystemid
        (string) - ID of the file system for whose tags will be modified.

    tags
        (dict) - The tags to add to the file system

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' boto_efs.create_tags
    """

    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)

    new_tags = []
    for k, v in tags.items():
        new_tags.append({"Key": k, "Value": v})

    client.create_tags(FileSystemId=filesystemid, Tags=new_tags)


def delete_file_system(
    filesystemid, keyid=None, key=None, profile=None, region=None, **kwargs
):
    """
    Deletes a file system, permanently severing access to its contents.
    Upon return, the file system no longer exists and you can't access
    any contents of the deleted file system. You can't delete a file system
    that is in use. That is, if the file system has any mount targets,
    you must first delete them.

    filesystemid
        (string) - ID of the file system to delete.

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' boto_efs.delete_file_system filesystemid
    """

    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)

    client.delete_file_system(FileSystemId=filesystemid)


def delete_mount_target(
    mounttargetid, keyid=None, key=None, profile=None, region=None, **kwargs
):
    """
    Deletes the specified mount target.

    This operation forcibly breaks any mounts of the file system via the
    mount target that is being deleted, which might disrupt instances or
    applications using those mounts. To avoid applications getting cut off
    abruptly, you might consider unmounting any mounts of the mount target,
    if feasible. The operation also deletes the associated network interface.
    Uncommitted writes may be lost, but breaking a mount target using this
    operation does not corrupt the file system itself.
    The file system you created remains.
    You can mount an EC2 instance in your VPC via another mount target.

    mounttargetid
        (string) - ID of the mount target to delete

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' boto_efs.delete_mount_target mounttargetid
    """

    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)

    client.delete_mount_target(MountTargetId=mounttargetid)


def delete_tags(
    filesystemid, tags, keyid=None, key=None, profile=None, region=None, **kwargs
):
    """
    Deletes the specified tags from a file system.

    filesystemid
        (string) - ID of the file system for whose tags will be removed.

    tags
        (list[string]) - The tag keys to delete to the file system

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' boto_efs.delete_tags
    """

    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)

    client.delete_tags(FileSystemId=filesystemid, Tags=tags)


def get_file_systems(
    filesystemid=None,
    keyid=None,
    key=None,
    profile=None,
    region=None,
    creation_token=None,
    **kwargs
):
    """
    Get all EFS properties or a specific instance property
    if filesystemid is specified

    filesystemid
        (string) - ID of the file system to retrieve properties

    creation_token
        (string) - A unique token that identifies an EFS.
        If fileysystem created via create_file_system this would
        either be explictitly passed in or set to name.
        You can limit your search with this.

    returns
        (list[dict]) - list of all elastic file system properties

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' boto_efs.get_file_systems efs-id
    """

    result = None
    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)

    if filesystemid and creation_token:
        response = client.describe_file_systems(
            FileSystemId=filesystemid, CreationToken=creation_token
        )
        result = response["FileSystems"]
    elif filesystemid:
        response = client.describe_file_systems(FileSystemId=filesystemid)
        result = response["FileSystems"]
    elif creation_token:
        response = client.describe_file_systems(CreationToken=creation_token)
        result = response["FileSystems"]
    else:
        response = client.describe_file_systems()

        result = response["FileSystems"]

        while "NextMarker" in response:
            response = client.describe_file_systems(Marker=response["NextMarker"])
            result.extend(response["FileSystems"])

    return result


def get_mount_targets(
    filesystemid=None,
    mounttargetid=None,
    keyid=None,
    key=None,
    profile=None,
    region=None,
    **kwargs
):
    """
    Get all the EFS mount point properties for a specific filesystemid or
    the properties for a specific mounttargetid. One or the other must be
    specified

    filesystemid
        (string) - ID of the file system whose mount targets to list
                   Must be specified if mounttargetid is not

    mounttargetid
        (string) - ID of the mount target to have its properties returned
                   Must be specified if filesystemid is not

    returns
        (list[dict]) - list of all mount point properties

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' boto_efs.get_mount_targets
    """

    result = None
    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)

    if filesystemid:
        response = client.describe_mount_targets(FileSystemId=filesystemid)
        result = response["MountTargets"]
        while "NextMarker" in response:
            response = client.describe_mount_targets(
                FileSystemId=filesystemid, Marker=response["NextMarker"]
            )
            result.extend(response["MountTargets"])
    elif mounttargetid:
        response = client.describe_mount_targets(MountTargetId=mounttargetid)
        result = response["MountTargets"]

    return result


def get_tags(filesystemid, keyid=None, key=None, profile=None, region=None, **kwargs):
    """
    Return the tags associated with an EFS instance.

    filesystemid
        (string) - ID of the file system whose tags to list

    returns
        (list) - list of tags as key/value pairs

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' boto_efs.get_tags efs-id
    """
    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)
    response = client.describe_tags(FileSystemId=filesystemid)
    result = response["Tags"]

    while "NextMarker" in response:
        response = client.describe_tags(
            FileSystemId=filesystemid, Marker=response["NextMarker"]
        )
        result.extend(response["Tags"])

    return result


def set_security_groups(
    mounttargetid,
    securitygroup,
    keyid=None,
    key=None,
    profile=None,
    region=None,
    **kwargs
):
    """
    Modifies the set of security groups in effect for a mount target

    mounttargetid
        (string) - ID of the mount target whose security groups will be modified

    securitygroups
        (list[string]) - list of no more than 5 VPC security group IDs.

    CLI Example:

    .. code-block:: bash

        salt 'my-minion' boto_efs.set_security_groups my-mount-target-id my-sec-group
    """

    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)
    client.modify_mount_target_security_groups(
        MountTargetId=mounttargetid, SecurityGroups=securitygroup
    )



"""
Pillar cache tests
"""

import pytest


@pytest.fixture()
def pillar_cache_tree(pillar_state_tree, pillar_salt_minion, pillar_salt_call_cli):
    top_file = """
    base:
      '{}':
        - test
    """.format(
        pillar_salt_minion.id
    )
    test_pillar = """
    test: one
    test2: two
    """
    top_tempfile = pytest.helpers.temp_file("top.sls", top_file, pillar_state_tree)
    pillar_tempfile = pytest.helpers.temp_file(
        "test.sls", test_pillar, pillar_state_tree
    )
    try:
        with top_tempfile, pillar_tempfile:
            ret = pillar_salt_call_cli.run("saltutil.refresh_pillar", wait=True)
            assert ret.returncode == 0
            assert ret.data is True
            yield
    finally:
        # Refresh pillar again to cleaup the temp pillar
        ret = pillar_salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True


@pytest.fixture()
def pillar_cache_tree_no_refresh(
    pillar_state_tree, pillar_salt_minion, pillar_salt_call_cli
):
    """
    setup the pillar tree but do not run saltutil.refresh_pillar
    after setting up pillar
    """
    ret = pillar_salt_call_cli.run("saltutil.refresh_pillar", wait=True)
    top_file = """
    base:
      '{}':
        - test
    """.format(
        pillar_salt_minion.id
    )
    test_pillar = """
    test: one
    test2: two
    test3: three
    """
    top_tempfile = pytest.helpers.temp_file("top.sls", top_file, pillar_state_tree)
    pillar_tempfile = pytest.helpers.temp_file(
        "test.sls", test_pillar, pillar_state_tree
    )
    try:
        with top_tempfile, pillar_tempfile:
            ret = pillar_salt_call_cli.run(
                "saltutil.refresh_pillar", wait=True, clean_cache=False
            )
            yield
    finally:
        # Refresh pillar again to cleaup the temp pillar
        ret = pillar_salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True


def test_pillar_cache_refresh(pillar_cache_tree, pillar_salt_call_cli):
    """
    Test pillar cache updates after a refresh_pillar
    """
    ret = pillar_salt_call_cli.run("pillar.items")
    assert ret.returncode == 0
    assert ret.data
    assert "test" in ret.data
    assert "test2" in ret.data


def test_pillar_cache_items(pillar_cache_tree_no_refresh, pillar_salt_call_cli):
    """
    Test pillar cache does not refresh pillar when using pillar.items
    """
    # pillar.items should be empty
    pillar_before = pillar_salt_call_cli.run("pillar.items").data
    pillar_salt_call_cli.run("saltutil.refresh_pillar", wait=True)
    # pillar.items should contain the new pillar data
    ret = pillar_salt_call_cli.run("pillar.items")
    assert ret.returncode == 0
    assert ret.data
    assert "test" in ret.data
    assert "test2" in ret.data
    assert ret.data != pillar_before


def test_pillar_cache_passes_extra_minion_data(pillar_salt_call_cli):
    """
    Test that pillar cache does not disable passing of
    extra_minion_data to external pillars
    """
    ret = pillar_salt_call_cli.run("pillar.items")
    assert ret.returncode == 0
    assert ret.data
    assert "hi" in ret.data
    assert ret.data["hi"] == "there"



"""
Tests for the spm files utility
"""

import os
import shutil

import pytest

from tests.support.case import SPMCase


@pytest.mark.windows_whitelisted
@pytest.mark.destructive_test
class SPMFilesTest(SPMCase):
    """
    Validate the spm files command
    """

    def setUp(self):
        self.config = self._spm_config()
        self._spm_build_files(self.config)

    @pytest.mark.slow_test
    def test_spm_files(self):
        """
        test spm files
        """
        self._spm_create_update_repo(self.config)
        install = self.run_spm("install", self.config, "apache")
        get_files = self.run_spm("files", self.config, "apache")

        os.path.exists(
            os.path.join(self.config["formula_path"], "apache", "apache.sls")
        )

    def tearDown(self):
        shutil.rmtree(self._tmp_spm)
